"""
Module for evaluating cubic tensor spline coefficients given the data
to be interpolated on a uniformly spaced tensor product grid.

Apologies for not writing this up in LaTeX. Hopefully nobody will ever need
to read it and the code will just work.

Suppose we have a d-dimensional grid with dimensions (N1 x N2 x ... x Nd).
Denote x_{i1, i2, ..., id} as the coordinates of the grid point with indices
[i1, i2, ..., id]. We are interested in interpolating some function f(x),
where x is a d-dimensional vector and f is real valued. Denote the grid data
by f_{i1, ..., id} = f(x_{i1, ..., id}), which is known.

A cubic tensor spline
interpolant for f requires a grid of ((N1 + 2) x (N2 + 2) x ... x (Nd + 2))
spline coefficients. We can extend the coordinate grid to a "coefficient grid"
by adding one additional grid point on both boundaries of each grid dimension.
We pad the grid data on these boundary points with zeros, and we will need to
impose boundary conditions to determine these coefficients.

The spline interpolant evaluated at x = (x1, ..., xd) is of the form
  TS(x) = \sum_{j1, ..., jd} c_{j1, ..., jd} s1_{j1}(x1) ... sd_{jd}(xd)
where the index ji runs over the (Ni + 2) spline coefficient indices,
c_{j1, ..., jd} are the spline coefficients, and si_{ji} is a one-dimensional
cubic BSpline associated with point ji in the j'th dimension of the coefficient
grid. Evaluating this on the spline coefficient grid gives us
((N1+2) x ... x (Nd+2)) equations:
    f_{i1, ..., id} =
        \sum_{j1, ..., jd} c_{j1, ..., jd} M1_{i1, j1} ... Md_{id, jd}
where Mk_{ik, jk} is a ((Nk+2) x (Nk+2)) matrix giving the BSpline associated
with point jk of the k'th dimension of the coefficient grid evaluated at the
point ik of the k'th dimension of the *padded* coordinate grid. When ik is one
of the boundary points (0 or Nk+1), we are free to set Mk_{ik, jk} as we wish
to control the boundary conditions which will result when we solve this system
of equations. 

The matrices Mk are exactly those used when solving for coefficients of a
one-dimensional cubic spline! For example, with 'not-a-knot' (constant third
derivative) boundary conditions, we have:

M= [[1,    -2,    3/2,   -2/3,    1/6,      0,      0,      0,      0],
    [1,     0,      0,      0,      0,      0,      0,      0,      0],
    [0,   1/4,   7/12,    1/6,      0,      0,      0,      0,      0],
    [0,     0,    1/6,    2/3,    1/6,      0,      0,      0,      0],
    [0,     0,      0,    1/6,    2/3,    1/6,      0,      0,      0],
    [0,     0,      0,      0,    1/6,    2/3,    1/6,      0,      0],
    [0,     0,      0,      0,      0,    1/6,   7/12,    1/4,      0],
    [0,     0,      0,      0,      0,      0,      0,      0,      1],
    [0,     0,      0,      0,    1/6,   -2/3,    3/2,     -2,      1]]

Here, the first and last rows represent the 'not-a-knot' boundary condition.
If we then look at column j, ignoring the first and last rows we have the
j'th BSpline evaluated at each of the Nk grid points.
The first three and last three columns differ from the middle (bulk) columns
because the spline breakpoints are not uniformly spaced near the boundary and
so the BSplines near the boundary are different.
For example, this (9x9) matrix corresponds to a grid of 7 uniformly spaced
points which we will take to be [0, 1, 2, 3, 4, 5, 6].
The spline breakpoints would then be [0, 0, 0, 0, 1, 2, 3, 4, 5, 6, 6, 6, 6].
The first spline coefficient corresponds to a cubic BSpline with breakpoints
[0, 0, 0, 0, 1], which is 0 outside of [0, 1), and rolls smoothly from 1 at x=0
to 0 at x=1. So this BSpline is 1 at the first grid point and 0 at all other
grid points, giving us the first column of the matrix (ignoring the first and
last rows). The second spline coefficient corresponds to a cubic BSpline with
breakpoints [0, 0, 0, 1, 2]. This BSpline 1/4 at x=1 and 0 at all other grid
points, giving us the second column (again neglecting the first and final rows).
The third column also corresponds to a boundary BSpline, but once we reach the
fourth column we have a BSpline with breakpoints [0, 1, 2, 3, 4] which is a
"typical" or "bulk" case - the breakpoints are uniformly spaced, and all "bulk"
BSplines are identical up to a translation. This BSpline has support [0, 4),
is 0 at the boundaries of its support, and takes on the values 1/6, 2/3, and 1/6
at x=1, x=2, and x=3 respectively.

Okay, now we are ready to solve our HUGE ((N1+2) x ... x (Nd+2)) linear system
of equations without too much pain. We first find inverses of the Mk matrices,
denoted MkInv, such that \sum_{i} MkInv{n, i} Mk_{i,j} = \delta_{n, j}.
We can then compute
    \sum_{i1, ..., id} f_{i1, ..., id} M1Inv{n1, i1} ... MdInv{nd, id} =
    \sum_{j1, ..., jd} c_{j1, ..., jd} \delta_{n1, j1} ... \delta{nd, jd} =
    c_{n1, ..., nd}
So we can solve for all the coefficients just by summing our padded grid data
against all of these (small) matrices! Of course, since we have many indices,
this requires some nice numpy tensordot magic :)
"""

import numpy as np

COEFS = {
    "bulk": [(-1, 1./6.), (0, 2./3.), (1, 1./6.)],
    "inner edge": [(-1, 1./4.), (0, 7./12.), (1, 1./6.)],
    "outer edge": [(-1, 1.)],
    "BC not-a-knot": [(0, 1.), (1, -2.), (2, 1.5), (3, -2./3.), (4, 1./6.)],
    "BC natural": [(0, 1.), (1, -1.5), (2, 0.5)],
        }

def get_1d_spline_matrix(n, bc="not-a-knot"):
    """
    Constructs the matrix A such that A*c = f, where c is a vector of spline
    coefficients and f is the solution evaluated on the knots, padded with one
    zero on either side for boundary conditions.

    n: The number of spline coefficients (number of knots + 2)
    bc: The boundary condition. Can be "not-a-knot" (constant 3rd derivative)
        or "natural".
    """
    if n < 6:
        raise Exception("Not implemented for < 4 knots, using cubic splines!")
    matrix = np.zeros((n, n))
    bc_coef_key = "BC %s"%(bc)
    if not bc_coef_key in COEFS.keys():
        raise Exception("Unknown boundary condition: %s"%(bc))

    # Start with the boundaries and work our way in
    for j, c in COEFS[bc_coef_key]:
        matrix[0, j] = c
        matrix[-1, -1-j] = c

    for j, c in COEFS["outer edge"]:
        matrix[1, 1+j] = c
        matrix[-2, -2-j] = c

    for j, c in COEFS["inner edge"]:
        matrix[2, 2+j] = c
        matrix[-3, -3-j] = c

    for i in range(3, n-3):
        for j, c in COEFS["bulk"]:
            matrix[i, i+j] = c
    return matrix


class UniformSpacingCubicSplineND:
    """
Computes coefficients for an N-dimensional cubic spline
on uniformly-spaced grid data.
    """

    def __init__(self, dimensions, BC='not-a-knot'):
        """
dimensions:
    (nx_1, nx_2, ...nx_d), number of points per dimension
BC:
    Boundary conditions for the splines. Can be 'not-a-knot' or 'natural'.
        """
        self.dims = dimensions
        self.d = len(dimensions)
        self.nCoefs = np.array([n+2 for n in dimensions])
        self.N = len(dimensions)
        self.nTotal = np.prod(self.nCoefs)
        self.BC = BC

        t_seconds = self.nTotal * 4.e-7 # ~ order of magnitude
        if t_seconds > .01:
            print('%s coefficients: solves should take O(%0.1e seconds) each.'%(
                self.nTotal, t_seconds))

        self.setup_1d_matrices()

    def setup_1d_matrices(self):
        """
Due to the tensor-product structure, we do not need to invert
the ((nx_1 * ... * nx_d) X (nx_1 * ... * nx_d)) system of
equations, and can instead solve a small system for each dimension.
Here we find the necessary inverse matrices.
        """
        self.inv_1d_matrices = []
        for n in self.nCoefs:
            matrix = get_1d_spline_matrix(n)
            # TODO: Better method for inverting?
            # These matrices are almost tridiagonal
            self.inv_1d_matrices.append(np.linalg.inv(matrix))

    def solve(self, griddata):
        """
Given the function evaluated on the knots, computes the tensor-spline
coefficients that can be used to interpolate the function.
        """
        if not np.shape(griddata) == self.dims:
            raise ValueError("griddata should have shape {}".format(self.dims))

        tmp_result = np.pad(griddata, 1, 'constant')

        # We will apply the 1d inverse matrices from last to first.
        # With each application, the shape of tmp_result goes from
        # (for example) (a, b, c, d) -> (d, a, b, c) -> (c, d, a, b)
        # so that we are always applying the 1d grid matrix to the last index.
        # We also end up with the correct shape at the end and don't have to
        # worry about multidimensional transposes.
        for minv in self.inv_1d_matrices[::-1]:
            tmp_result = np.tensordot(minv, tmp_result, (1, self.d - 1))

        return tmp_result



# --- gwtools.py ---

"""
	A collection of useful gravitational wave tools
"""

from __future__ import division # for python 2


__copyright__ = "Copyright (C) 2014 Scott Field and Chad Galley"
__email__     = "sfield@astro.cornell.edu, crgalley@tapir.caltech.edu"
__status__    = "testing"
__author__    = "Scott Field, Chad Galley"

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

import numpy as np
from scipy.interpolate import interp1d
from scipy.optimize import minimize


# Define global constants
try: # only used when evaluating lal routines. TODO: move to lal-specific file
  from lal import PC_SI
  from lal import MSUN_SI
  from lal import MTSUN_SI as Msuninsec
except: # ... or use gwtools constants for consistency?
  from . import const
  import warnings as _warnings
  G = const.G
  c = const.c
  PC_SI   = const.Mpcinkm*1000.0 # (in meters = Mpcinkm)
  MSUN_SI = const.Msun
  Msuninsec = const.Msuninsec
  #raise Warning("using non-LAL constants")
  _warnings.warn("using non-LAL constants")


#####################################
# Functions for changing parameters #
#####################################

def m1m2_to_Mc(m1,m2):
    """Chirp mass from m1, m2"""
    return (m1*m2)**(3./5.)/(m1+m2)**(1./5.)


def m1m2_to_nu(m1,m2):
    """Symmetric mass ratio from m1, m2"""
    return m1*m2/(m1+m2)**2


def m1m2_to_Mcnu(m1, m2):
  """Compute symmetric mass ratio and chirp mass from m1, m2"""	
  return m1m2_to_Mc(m1,m2), m1m2_to_nu(m1,m2)


def q_to_nu(q):
  """Convert mass ratio (which is >= 1) to symmetric mass ratio"""
  return q / (1.+q)**2.


def nu_to_q(nu):
  """Convert symmetric mass ratio to mass ratio (which is >= 1)"""
  return (1.+np.sqrt(1.-4.*nu)-2.*nu)/(2.*nu)


def Mq_to_m1m2(M, q):
  """Convert total mass, mass ratio pair to m1, m2"""
  m2 = M/(1.+q)
  m1 = M-m2
  return m1, m2


def Mq_to_Mc(M, q):
  """Convert mass ratio, total mass pair to chirp mass"""
  return M*q_to_nu(q)**(3./5.)


def Mcq_to_M(Mc, q):
  """Convert mass ratio, chirp mass to total mass"""
  return Mc*q_to_nu(q)**(-3./5.)


def Mcnu_to_M(Mc, nu):
  """Convert chirp mass and symmetric mass ratio to total mass"""
  return Mc*nu**(-3./5.)

def Mnu_to_Mc(M, nu):
  """Convert total mass and symmetric mass ratio to chirp mass"""
  return M*nu**(3./5.)

def Mnu_to_m1m2(M, nu):
  """Convert total mass and symmetric mass ratio to m1, m2"""
  q      = nu_to_q(nu)
  m1, m2 = Mq_to_m1m2(M,q) # m1 > m2. m1 and m2 will be SI
  return m1, m2

def Mcnu_to_m1m2(Mc, nu):
  """Convert chirp mass, symmetric mass ratio pair to m1, m2"""
  q = nu_to_q(nu)
  M = Mcq_to_M(Mc, q)
  return Mq_to_m1m2(M, q)

def m1m2_to_delta(m1, m2):
  """Convert m1, m2 pair to relative mass difference [delta = (m1-m2)/(m1+m2)]"""
  return (m1-m2)/(m1+m2)


def q_to_delta(q):
  """Convert mass ratio (which is >= 1) to relative mass difference (delta)"""
  return (q-1.)/(q+1.)


def delta_to_q(delta):
  """Convert relative mass difference (delta) to mass ratio q (q >= 1)"""
  return (1.+delta)/(1.-delta)


def delta_to_nu(delta):
  """Convert relative mass difference (delta) to symmetric mass ratio"""
  return (1.-delta**2)/4.


def nu_to_delta(nu):
  """Convert symmetric mass ratio to relative mass difference delta"""
  return np.sqrt(1.-4.*nu)


def X1X2_to_Xs(X1, X2):
  """Convert dimensionless spins X1, X2 to symmetric spin Xs"""
  return (X1+X2)/2.


def X1X2_to_Xa(X1, X2):
  """Convert dimensionless spins X1, X2 to anti-symmetric spin Xa"""
  return (X1-X2)/2.


def X1X2_to_XsXa(X1, X2):
  """Convert dimensionless spins X1, X2 to symmetric and anti-symmetric spins Xs, Xa"""
  return X1X2_to_Xs(X1,X2), X1X2_to_Xa(X1,X2)

def Chi1M1Chi2M2_to_chiE(chi1,m1,chi2,m2):
  """Convert (chi_i,m_i) pair to chi_effective"""
  M = m1+m2
  return ( m1 * chi1 + m2 * chi2) / M

##################################
# Miscellaneous helper functions #
##################################

def fgwisco(Mtot):
  """GW frequency at ISCO. [Note: Maggiore's text has an extra 1/2.]"""
  return 6.0**(-1.5) / (np.pi*Mtot)


def get_arg(a, a0):
  """Get argument at which a0 occurs in array a"""
  return np.argmin(np.abs(a-a0))


def get_peak(x, y):
  """Get argument and values of x and y at maximum value of |y|"""
  arg = np.argmax(y)
  return [arg, x[arg], y[arg]]


def all_zero_crossings(x):
  pos = x > 0
  npos = ~pos
  return ((pos[:-1] & npos[1:]) | (npos[:-1] & pos[1:])).nonzero()[0]


def fit_peak(x, y, fit='polyfit', a=10, b=10, deg=2, num=1000):
  """
  Estimate peak of y through interpolation.
  
  Input
  =====
  x   --- x data array (1d)
  y   --- y data array (1d)
  fit --- Type of fitting function 
          (default='polyfit')
  a   --- Number of points to fit before estimated peak
          (default=10)
  b   --- Number of points to fit after estimated peak 
          (default=10)
  deg --- Degree or order of fitting function 
          (default=2)
  num --- Number of high resolution x samples for estimating peak 
          (default=1000)
  
  Output
  ======
  x_peak --- Estimated x value of the estimated maximum of y data
  y_peak --- Estimated maximum of y data
  
  Note: The fitting is currently done with the UnivariateFits class 
  in ROMpy. The maximum of the y data is estimated by evaluating the 
  fitting function at num(=1000) points. However, the default fitting 
  function is a quadratic polynomial, the peak of which is 
  calculated analytically instead of numerically.
  """
  
  try:
    from rompy.univar import UnivariateFits
  except:
    raise Exception("Cannot import UnivariateFits class from RomPy.")
  
  # Get argument of discrete data y
  try:
    from rompy.derivatives import D
  except:
    raise Exception("Cannot import D function for computing finite differences from RomPy.")

  arg = np.argmax(y)
  if arg-a < 0:
    a = arg
  if arg+b > len(y):
    b = 0
  x_fit = x[arg-a:arg+b+1]
  y_fit = y[arg-a:arg+b+1]
  
  # Determine if the data in this range has a maximum
  dy = D(y_fit, x_fit, dx=1, order=4)
  d2y = D(y_fit, x_fit, dx=2, order=4)
  i0 = all_zero_crossings(dy)
  if len(i0) == 0:
    raise Exception("Did not find an extremum in the input data except possibly at endpoints.")
  elif len(i0) > 1:
    raise Exception("Found more than one extremum in the input data.")
  elif len(i0) == 1:
    # Positive concavity implies a minimum
    if d2y[i0] > 0.:
      raise Exception("Did not find a maximum in the input data.")
    # Zero concavity implies an inflection point
    elif d2y[i0] == 0.:
      raise Exception("Extremum is an inflection point.")
    # Negative concavity implies a maximum, so continue on
  else:
    raise Exception("Cannot determine the nature of the data. Please check the input.")
  
  # Fit over the interval around arg using the requested fitting function
  _fit = UnivariateFits(x_fit, y_fit, fit=fit, deg=deg)
  if fit == 'polyfit' and deg == 2:
    c, b, a = _fit._eval
    x_peak = -b/(2.*c)
  else:
    x_fine = np.linspace(x_fit[0], x_fit[-1], num)
    x_peak = x_fine[np.argmax(_fit.eval(x_fine))]
  
  y_peak = _fit.eval(x_peak)
  return x_peak, y_peak


def chop(x, y, xmin, xmax):
  """Chop arrays x, y for x in [xmin, xmax]"""
  argmin = get_arg(x, xmin)
  argmax = get_arg(x, xmax)
  return x[argmin:argmax+1], y[argmin:argmax+1]


def shift(x, xshift):
  """Shift array x by xshift"""
  arg = get_arg(x, xshift)
  return x-x[arg]


def shift_chop(x, y, xshift, xmin, xmax):
  """Shift x, y arrays by xshift then chop for x in [xmin, xmax]"""
  xnew = shift(x, xshift)
  xminnew = xmin-xshift
  xmaxnew = xmax-xshift
  #return chop(xnew, y, xminnew, xmaxnew)
  return chop(xnew, y, xmin, xmax)


def amp(h):
  """Get amplitude of waveform, h = A*exp(i*phi)"""
  return np.abs(h)


def amp_phase(h):
  """Get amplitude and phase of waveform, h = A*exp(i*phi)"""
  amp = np.abs(h);
  return amp, phase(h)


def phase(h):
  """Get phase of waveform, h = A*exp(i*phi)"""
  if np.shape(h):
    # Compute the phase only for non-zero values of h, otherwise set phase to zero.
    nonzero_h = h[np.abs(h) > 1e-300]
    phase = np.zeros(len(h), dtype='double')
    phase[:len(nonzero_h)] = np.unwrap(np.real(-1j*np.log(nonzero_h/np.abs(nonzero_h))))
  else:
    nonzero_h = h
    phase = np.real(-1j*np.log(nonzero_h/np.abs(nonzero_h)))
  return phase


def cycles(h):
  """Count number of cycles (to merger, if present) in waveform"""
  phi = phase(h)
  ipk, phi_pk, A_pk = get_peak(phi, np.abs(h))
  return (phi_pk - phi[0])/(2.*np.pi)


def hplus(h):
  """Get real part of waveform, h = A*exp(i*phi)"""
  return np.real(h)


def hcross(h):
  """Get imaginary part of waveform, h = A*exp(i*phi)"""
  return np.imag(h)


def logd(t,f):
  """Compute a logarithmic derivative"""
  dfdt = np.diff(f)/np.diff(t)
  return t[1:], dfdt/f[1:]


def plot_pretty(time, y, fignum=1, flavor='linear', color='k', linestyle=['-', '--'], \
        label=['$h_+(t)$', '$h_\\times(t)$'], legendQ=True, showQ=False):
  """create a waveform figure with nice formatting and labels.
  returns figure method for saving, plotting, etc."""

  try:
    import matplotlib.pyplot as plt
  except:
    raise Exception("Cannot import matplotlib.")
  
  # Plot waveform
  fig = plt.figure(fignum)
  ax = fig.add_subplot(111)
  
  dim_y = np.shape(y)
  num_y = len(dim_y)
  num_color, num_label = np.size(color), np.size(label)
  dim_linestyle = np.shape(linestyle)
  
  if num_y > 2:
    raise ValueError("Can only plot one or two functions")
  
  for ii in range(num_y):
    
    if num_y == 1: yy = y
    else: yy = y[ii]
    
    if num_color == 1: cc = color
    else: cc = color[ii]
    
    if num_label == 1: ll = label
    else: ll = label[ii]
    
    if len(dim_linestyle) == 0: ss = linestyle
    elif len(linestyle) == 1: ss = linestyle[0]
    else: ss = linestyle[ii]
    
    if flavor == 'linear':
      if legendQ:
        plt.plot(time, yy, color=cc, linestyle=ss, label=ll)
      else:
        plt.plot(time, yy, color=cc, linestyle=ss)
      
    elif flavor == 'semilogy':
      if legendQ:
        plt.semilogy(time, yy, color=cc, linestyle=ss, label=ll)
      else:
        plt.semilogy(time, yy, color=cc, linestyle=ss)
        
    elif flavor == 'semilogx':
      if legendQ:
        plt.semilogx(time, yy, color=cc, linestyle=ss, label=ll)
      else:
        plt.semilogx(time, yy, color=cc, linestyle=ss)
    
    elif flavor == 'loglog':
      if legendQ:
        plt.loglog(time, yy, color=cc, linestyle=ss, label=ll)
      else:
        plt.loglog(time, yy, color=cc, linestyle=ss)
    
    else:
      raise ValueError("Not a valid plot type")
  
  plt.xlabel('Time')
  plt.ylabel('Waveform')
  
  if legendQ:
    plt.legend(loc='upper left')
  
  if showQ:
    plt.show()
  
  return fig


def ecc_estimator(t,h,fit_window,type=1,fit_deg=1):
  """"Estimate the eccentricity associated with gravitational 
  waveform h from Eq. (17) of arxiv:1004.4697 (gr-qc)."""
    
  ### TODO: add butterworth filter for fit residuals. what would amplitude estimator look like?
    
  ### crop h to fit interval ###
  ecc_interval = np.arange(fit_window[0],fit_window[1])
  t            = t[ecc_interval]
  h            = h[ecc_interval]
  
  if type == 1:
    
    ### fit phase with degree fit_deg polynomial on fit_window ###
    amp_tmp, phase = get_amp_phase(h)
    p_coeff        = np.polyfit(t,phase,fit_deg)
    phase_fit      = np.polyval(p_coeff,t)
    
    ### compute the estimator ###
    ecc_est = ( phase - phase_fit ) / 4.
    
    return t, ecc_est
    
  elif type == 2:
    
    amp_tmp        = amp(h)
    t, amp_log     = logd(t,amp_tmp) # t will be one element shorter than before
    p_coeff        = np.polyfit(t,amp_log,fit_deg)
    amplog_fit     = np.polyval(p_coeff,t)
    
    return t, (amp_log - amplog_fit)


def find_instant_freq(hp, hc, t):
  """instantaneous starting frequency for 

              h = A(t) exp(2 * pi * i * f(t) * t), 

      where we approximate \partial_t A ~ \partial_t f ~ 0."""

  h    = hp + 1j*hc
  dt   = t[1] - t[0]
  hdot = (h[2] - h[0]) / (2 * dt) # 2nd order derivative approximation at t[1]

  f_instant = hdot / (2 * np.pi * 1j * h[1])
  f_instant = f_instant.real

  return f_instant


def dimensionless_time(M,t,UnitsOfM='s'):
  """ input time and mass in seconds. Returns dimensionless time """
  # TODO: input M could be in solar masses or 
  if UnitsOfM=='s':
    return t/M
  else:
    raise ValueError("Unknown units of M")

def dimensionless_frequency(M,f):
  """ Input frequency in seconds, Mass in solar masses. Returns dimensionless frequency"""
  return f * Msuninsec * M

def remove_amplitude_zero(t,h):
  """ removes h[i] t[i] from array if |h[i]| = 0 """
  
  amp, phase     = amp_phase(h)
  where_non_zero = np.nonzero(amp)
  
  return t[where_non_zero], h[where_non_zero]


def modify_phase(h,offset):
  """ Modify GW mode's phase to be \phi(t) -> \phi(t) + offset.
      For h_{ell,m}, typically offset = m*z_rot where z_rot is a 
      physical rotation about the z-axix (orthogonal to the orbital plane)."""
  return  h*np.exp(1.0j * offset)


def coordinate_time_shift(t,offset):
  """ Modify times to be t -> t + offset """
  return t + offset


def find_common_time_window(t1,t2):
  """ Given two temporal grids, find the largest range of common times 
      defined by [min_common,max_common] """

  min_common = max( t1[0], t2[0] )
  max_common = min( t1[-1], t2[-1] )
  
  if (max_common <= min_common):
    raise ValueError("there is no common time grid")
  
  return min_common, max_common


def simple_align_params(t1,h1,t2,h2):
  """ t1 times for complex waveform h1. 
  
      This routine returns simple alignment parameters 
      deltaT and deltaPhi by...
        (i)  fining discrete waveform peak
        (ii) aligning phase values at the peak time """
  
  amp1,phase1 = amp_phase(h1)
  amp2,phase2 = amp_phase(h2)
  
  deltaT   = t1[np.argmax( amp1 )] - t2[np.argmax( amp2 )]
  deltaPhi = phase1[np.argmax( amp1 )] - phase2[np.argmax( amp2 )]
  
  return deltaT, deltaPhi


def euclidean_norm_sqrd(f,dx):
  """ Euclidean norm squared of a complex vector f """
  return (np.sum(f*np.conj(f)) * dx).real

def euclidean_rel_error_sqrd(f1,f2,dx):
  """ Relative error 

            || f2 - f1||^2 / ( ||f1|| * ||f2|| )

computed from complex vectors f1 and f2 """

  err     = euclidean_norm_sqrd(f2-f1,dx)
  norm_f1 = np.sqrt(euclidean_norm_sqrd(f1,dx))
  norm_f2 = np.sqrt(euclidean_norm_sqrd(f2,dx))

  return err/(norm_f1*norm_f2)

def euclidean_norm_sqrd_2sphere(f,dx):
  """ Euclidean norm squared of a complex function 
  
      h(t,theta,phi) = \sum h_{ell,m}(t) B_{ell,m} (theta,phi) 

      known through its harmonic coefficients h_{ell,m}, each
      of which are column vector of an input matrix f."""

  full_norm = 0.0
  for ii in range(f.shape[1]):
    full_norm += euclidean_norm_sqrd(f[:,ii],dx)
    
  return full_norm


def generate_parameterize_waveform(t,h1_func,h1_type,h1_params=None):
  """ returns function for h(t;tc,phic) evaluations.

  INPUT
  =====
  t         --- array of times
  h1_func   --- waveform which can be evaluated for input set of times 
  h1_type   --- h1_func type ('interp1d','h_sphere') 
  h1_params --- any additional parameters needed for waveform evalautions"""

  if h1_type == 'interp1d': # interpolant built from scipy.interpolate.interp1d(t,h1_data)
    def parameterize_waveform(x):
      tc   = x[0]
      phic = x[1]
      h1_eval = h1_func( coordinate_time_shift(t,tc) ) # differing sign from minimize_norm_error is correct
      h1_eval = modify_phase(h1_eval,-phic)
      return h1_eval
      
  elif h1_type == 'h_sphere': # waveform from h_sphere_builder
    theta = h1_params
    def parameterize_waveform(x):
      tc   = x[0]
      phic = x[1]
      times = coordinate_time_shift(t,tc)
      hp,hc = h1_func(times,theta=theta, phi=0.0, z_rot=phic, psi_rot=None)
      return hp + 1.0j*hc

  else:
    raise ValueError('unknown waveform type')

  return parameterize_waveform


def generate_parameterized_norm(h1,h_ref,mynorm,t=None):
  """ 
  this routine will return a parameterized discrete norm 

         N(p1,p2,...) = || h1(p1,p2,...) - h_ref || 

  which can be minimized over the parameters (p1,p2...). If p1 = p2 = ... = 0 then 
  the norm is simply

                 || h1 - h_ref ||   for h1 = h1[t]

  h_ref is a reference waveform (represented by a discrete set
  evaluations) to which we match another waveform h1 (represented as a
  function). 

  h1 is a python function which takes inputs p1, p2,...
  where, typically, p1 is timeshift (tc) and p2 is a rotation about 
  the z-axis phic. The function h1 returns a vector h1(t_c,phi_c) collocated
  with h2_ref.


  Input
  =====
  t:         array of times such that h2_ref = h_ref[t]
  h2_ref:    array of reference waveform evaluations
  h1:        parameterized waveform function 
  mynorm:    function s.t. mynorm(f,dt) is a discrete norm


  Input expectations
  ==================
  (i) h1 should be defined on a larger temporal grid 
      than t and, hence, h2_ref. Why? When looking for the 
      minimum, h1 will be evaluated at times t + deltaT. 
      t should be viewed as the "common set of times" on which both 
      h1 and h_ref are known. """

  dt = 1.0 # we optimize the relative errors, factors of dt cancel

  # TODO: should check for TypeError by seeing if parameterized waveform can be evaluated by passing single x
  def ParameterizedNorm(x):
    h1_trial = h1(x)
    diff_h = h1_trial - h_ref
    
    # normalize by h2_ref as its fixed. Goal is to match h1 to h_ref #
    # TODO: perhaps divide by norm of h1_trial too, or pass in more general function to this routine
    overlap_errors = mynorm(diff_h,dt)/mynorm(h_ref,dt) 
    
    return overlap_errors

  return ParameterizedNorm


def create_common_time_grid(t1,t2,t_low_adj,t_up_adj):
  """ from temporal grids t1 and t2, create grid of common times"""

  common_dt      = (t1[2] - t1[1]) # TODO: t2 or t1 or variable
  t_start, t_end = find_common_time_window(t1,t2)
  common_times   = np.arange(t_start+t_low_adj,t_end-t_up_adj,common_dt) # small buffer needed 

  return common_times, t_start, t_end, common_dt


def discrete_minimization_from_discrete_waveforms(t,h,t_ref,h_ref):
  deltaT, deltaPhi = simple_align_params(t,h,t_ref,h_ref)
  h                = modify_phase(h,-deltaPhi)
  t                = coordinate_time_shift(t,-deltaT) # different sign from generate parameterize norm is correct

  return deltaT, deltaPhi, t, h


def setup_minimization_from_discrete_waveforms(t1,h1,t2,h2,t_low_adj,t_up_adj,verbose=False):

  t1, h1 = remove_amplitude_zero(t1,h1)
  t2, h2 = remove_amplitude_zero(t2,h2)

  if( (t1[-1] - t1[0]) < (t2[-1] - t2[0]) ):
    raise ValueError('first waveform should be longer')


  deltaT, deltaPhi, t1, h1 = discrete_minimization_from_discrete_waveforms(t1,h1,t2,h2)

  common_times, t_start, t_end, common_dt = create_common_time_grid(t1,t2,t_low_adj,t_up_adj)

  h1_interp = interp1d(t1,h1)
  h2_interp = interp1d(t2,h2)

  h2_eval = h2_interp(common_times)

  if(verbose):
    common_times_full = np.arange(t_start,t_end,common_dt)
    h2_eval_full_nrm  = mynorm(h2_interp(common_times_full),1.0)
    h2_eval_nrm       = mynorm(h2_interp(common_times),1.0)
    print("|| h_full || /  || h_adjmynorm(h2_eval_full) = ",h2_eval_full_nrm/h2_eval_nrm)

  return h1_interp, h2_eval, common_times, deltaT, deltaPhi


def minimize_waveform_match(h1_parameterized,href,mynorm,start_values,method):
  """ write me: should pass vector of values to h1... will generalize """
  
  ParameterizedNorm = generate_parameterized_norm(h1_parameterized,href,mynorm)
  
  opt_result   = minimize(ParameterizedNorm, start_values, method=method,tol=1e-12)
  opt_solution = opt_result.x
  
  min_norm     = ParameterizedNorm(opt_solution) # norm's value at global minimum
  h1_align     = h1_parameterized(opt_solution) # h1 waveform optimally matched to href
  guessed_norm = ParameterizedNorm(start_values) # norm's value using initial guess
  
  return [guessed_norm,min_norm], opt_solution, h1_align


#TODO: better description needed
def minimize_norm_error(t1,h1,tref,href,mynorm,t_low_adj=.1,t_up_adj=.1,method='nelder-mead',verbose=False):
  """
  Input
  ===== 
  t1,h1:      time/waveform vectors sampled at equally spaced times
  tref,href:  a pair of reference time/waveform vectors
  mynorm:     norm function (e.g. euclidean_norm_sqrd(f,dx) ) which takes a vector f
  t_low,t_up: adjusments  to "clip" the start and end portions of the pre-aligned waveforms

  Output
  ====== 
  guessed:      relative norm error with discrete "guess" for tc and phic offsets 
  min_norm:     minimized norm error by solving optimization problem
  tc, phic:     time/phase offsets which solve the 2D minimization problem
  common_times: WRITE ME
  h1_align:     WRITE ME
  href_eval:      WRITE ME

  Output in form [guessed, min_norm]  [tc, phic]  [common_times, h1_align, href_eval]


  Input expectations
  ================== 
  (i)  t1 and tref should be equally spaced grid of times. 
  (ii) for waveforms of different length, (t1,h1) pair should be longer

  Output caveats
  ============== 
  (i) evaluating the norm with values of (tc,phic) might give slightly different answers
      depending on the order of shifts/interpolants etc."""

  ### from discrete waveform data (t1,h1) and (tref,href), return items needed to solve minimization problem ###
  h1_interp, href_eval, common_times, deltaT, deltaPhi = \
      setup_minimization_from_discrete_waveforms(t1,h1,tref,href,t_low_adj,t_up_adj,verbose)
  
  ### h1_parameterized returns evaluations at common_times for given (tc,phic), induces parameterized norm ###
  h1_parameterized  = generate_parameterize_waveform(common_times,h1_interp,'interp1d')
  
  [guessed_norm,min_norm], opt_solution, h1_align = minimize_waveform_match(h1_parameterized,href_eval,mynorm,[0.0,0.0],method)
  
  tc         = opt_solution[0] + deltaT
  phic       = opt_solution[1] + deltaPhi
  
  return [guessed_norm, min_norm], [tc, phic], [common_times, h1_align, href_eval]


def minimize_norm_error_multi_mode(t1,h1,tref,href,mode_list,ell_m_match,t_low_adj=.1,t_up_adj=.1,method='nelder-mead',verbose=False):
  """ performs single-mode match, uses optimal match parameters for multi-mode waveform alignment"""

  ### solve single mode match here ###
  ell_match = ell_m_match[0]
  m_match   = ell_m_match[0]
  
  mode_indx = mode_list.index((ell_match,m_match))
  h_ref_22 = href[:,mode_indx]
  h1_22    = h1[:,mode_indx]
  
  [guessed_norm_1mode, minimized_norm_1mode], [tc, phic], [common_times,h1_align,href_align] = \
    minimize_norm_error(t1,h1_22,tref,h_ref_22,euclidean_norm_sqrd,method='nelder-mead')
  z_rot_opt = -phic/m_match
  
  ### use single mode results for multi-mode waveform ###
  h_sphere = h_sphere_builder(mode_list,href.real,href.imag,tref)
  href_eval_hp, href_eval_hc    = h_sphere(common_times)
  href_eval = href_eval_hp + 1.0j*href_eval_hc
  
  h_sphere = h_sphere_builder(mode_list,h1.real,h1.imag,t1)
  h1_align_hp, h1_align_hc  = h_sphere(times=common_times+tc,z_rot=z_rot_opt)
  h1_align = h1_align_hp + 1.0j*h1_align_hc
  
  ### check multi-mode waveform's closeness ###
  h_diff = href_eval - h1_align
  min_norm_sphere = euclidean_norm_sqrd_2sphere(h_diff,1.0)/euclidean_norm_sqrd_2sphere(href_eval,1.0)
  
  ### compute errors mode-by-mode ###
  rel_mode_errors = compute_many_mode_errors(h1_align,href_eval,mode_list,euclidean_norm_sqrd)
  
  return [rel_mode_errors, min_norm_sphere], [tc, z_rot_opt], [common_times,h1_align,href_eval]


def h_sphere_builder(modes,hp,hc,t):
  """Returns a function h(t,theta,phi;z_rot,tc). This function can be evaluated 
     for rotations about z-axis and returns either (i) a list of modes or (ii) 
     evaluation on sphere at (theta, phi)

      INPUT
      =====
      modes --- array of modes (ell,m)
      hp/hc --- matrix modes. each column is a mode evaluation for (ell[i],m[i]) in modes
      t     --- arrary of times at which modes have been evaluated"""

  from .harmonics import sYlm as sYlm
  from scipy.interpolate import splrep
  from scipy.interpolate import splev

  ### fill dictionary with model's modes as a spline ###
  hp_modes_spline = dict()
  hc_modes_spline = dict()
  ii = 0
  for ell_m in modes:
    hp_modes_spline[ell_m] = splrep(t, hp[:,ii], k=3)
    hc_modes_spline[ell_m] = splrep(t, hc[:,ii], k=3)
    ii += 1

  ### time interval for valid evaluations ###
  t_min = t.min()
  t_max = t.max()

  ### create function which can be used to evaluate for h(t,theta,phi) ###
  def h_sphere(times,theta=None,phi=None,z_rot=None,psi_rot=None):
    """ evaluations h(t,theta,phi), defined as matrix of modes, or sphere evaluations.

        INPUT
        =====
        times     --- numpy array of times to evaluate at
        theta/phi --- angle on the sphere, evaluations after z-axis rotation
        z_rot     --- rotation angle about the z-axis (coalescence angle)
        psi_rot   --- overall phase adjustment of exp(1.0j*psi_rot) mixing h+,hx"""

    # TODO: restore this after testing
    #if times.min() < t_min or times.max() > t_max:
    #  raise ValueError('surrogate cannot be evaluated outside of its time window')

    if psi_rot is not None:
      raise ValueError('not coded yet')

    ### output will be h (if theta,phi specified) or hp_modes, hc_modes ###
    if theta is not None and phi is not None:
      h = np.zeros((times.shape[0],),dtype=complex)
    else:
      hp_modes = np.zeros((times.shape[0],len(modes)))
      hc_modes = np.zeros((times.shape[0],len(modes)))

    ### evaluate modes at times ###
    jj=0
    for ell_m in modes:

      hp_modes_eval = splev(times, hp_modes_spline[ell_m])
      hc_modes_eval = splev(times, hc_modes_spline[ell_m])

      ### apply rotation about z axis and evaluation on sphere if requested ###
      h_modes_eval  = hp_modes_eval + 1.0j*hc_modes_eval
      if z_rot is not None:
        h_modes_eval = modify_phase(h_modes_eval,z_rot*ell_m[1])

      if theta is not None and phi is not None:
        sYlm_value =  sYlm(-2,ll=ell_m[0],mm=ell_m[1],theta=theta,phi=phi)
        h = h + sYlm_value*h_modes_eval

      else:
        hp_modes[:,jj] = h_modes_eval.real
        hc_modes[:,jj] = h_modes_eval.imag

      jj+=1

    if theta is not None and phi is not None:
      return h.real, h.imag
    else:
      return hp_modes, hc_modes

  return h_sphere


def compute_many_mode_errors(h,h_ref,mode_list,mynorm):
  """input: matrix of modes h, h_ref, mode_list, and norm 
    output: dictionary relating mode to relative errors h_{ell,m} compared with href_{ell,m} """

  h_mode_diff = h_ref - h
  relative_mode_errs = dict()
  ii=0
  for ell_m in mode_list:
    err = mynorm(h_mode_diff[:,ii],1.0)/mynorm(h_ref[:,ii],1.0)
    relative_mode_errs[ell_m] = err
    ii += 1

  return relative_mode_errs


def interpolant_h(t, h, deg=3):
  """Compute spline interpolant of input data h at samples t."""
  
  from scipy.interpolate import splrep
  from scipy.interpolate import splev
  dtype = h.dtype
  if dtype == 'complex':
    interp_real = splrep(t, np.real(h), k=deg)
    interp_imag = splrep(t, np.imag(h), k=deg)
    return interp_real, interp_imag
  elif dtype == 'double':
    interp = splrep(t, h, k=deg)
    return interp
  else:
    raise Exception("Function to be interpolated must be real or complex.")
  pass


def interpolate_h(tin, h, tout, deg=3):
  """Interpolate input data h at samples tout."""

  from scipy.interpolate import splrep
  from scipy.interpolate import splev
  
  if tout[0] < tin[0] or tout[-1] > tin[-1]:
    print(">>> Warning: Requested samples are outside of interpolated domain.")
  
  dtype = h.dtype
  if dtype == 'complex':
    interp_real, interp_imag = interpolant_h(tin, h, deg=deg)
    return splev(tout, interp_real) + 1j*splev(tout, interp_imag)
  elif dtype == 'double':
    interp = interpolant_h(tin, h, deg=deg)
    return splev(tout, interp)
  else:
    raise Exception("Function to be interpolated must be real or complex.")
  pass


def spline_f(tin, h, deg=3):
  """ Return a function f_sp such that f_sp(t) returns the spline
      evaluation of f"""

  from scipy.interpolate import splrep
  from scipy.interpolate import splev

  dtype = h.dtype
  if dtype == 'complex':
    interp_real, interp_imag = interpolant_h(tin, h, deg=deg)
  elif dtype == 'double':
    interp = interpolant_h(tin, h, deg=deg)
  else:
    raise Exception("Function to be interpolated must be real or complex.")

  def f_sp(t):
    if type(t) == float or type(t) == int or type(t) == np.float64:
      if t < tin[0] or t > tin[-1]:
        print(">>> Warning: Requested sample value %f is outside of interpolated domain."%t)
    else:
      if t[0] < tin[0] or t[-1] > tin[-1]:
        print(">>> Warning: Requested samples are outside of interpolated domain.")

    if dtype == 'complex':
      return splev(t, interp_real) + 1j*splev(t, interp_imag)
    elif dtype == 'double':
      return splev(t, interp)

  return f_sp


##############################################
# Convenient interfaces for LALsim functions #
##############################################

#def EOBNRv2_LAL(Mtot,q,dt,fmin,Dist,inclination):
#    """ Dist in megaparsecs"""
#    
#    Dist = Dist * 1e6 * lal.PC_SI
#    M1,M2  = gwtools.Mq_to_m1m2(Mtot,q)
#    hp, hc = lalsim.SimInspiralChooseTDWaveform(phiRef=0.0,deltaT=dt,m1=M1,m2=M2,s1x=0.0,s1y=0.0,s1z=0.0,\
#                                                s2x=0.0,s2y=0.0,s2z=0.0,f_min=fmin,f_ref=0.0,\
#                                                r=Dist,i=inclination,lambda1=0.0,lambda2=0.0,\
#                                                waveFlags=None,nonGRparams=None,amplitudeO=0,phaseO=7,\
#                                                approximant=lalsim.EOBNRv2)
#    h      = hp.data.data + (1j)*hc.data.data
#    times  = np.arange(np.size(hp.data.data))*hp.deltaT
#    
#    times,h = gwtools.remove_amplitude_zero(times,h) # removes end portion of waveform which is zero
#    
#    return times, h


def EOBNRv2_LAL(Mtot,q,dt,fmin,Dist,inclination):
  """ Simplified interface to EOBNRv2 (dominant mode). Dist in megaparsecs"""

  from lalsimulation import SimIMREOBNRv2DominantMode

  Dist = Dist * 1e6 * PC_SI
  Mtot = Mtot * MSUN_SI

  M1,M2 = Mq_to_m1m2(Mtot,q)
    
  hp, hc = SimIMREOBNRv2DominantMode(phiC=0.0,deltaT=dt,m1SI=M1,m2SI=M2,\
                                     fLower=fmin,distance=Dist,inclination=inclination)
    
  h        = hp.data.data + (1j)*hc.data.data
  times    = np.arange(np.size(hp.data.data))*hp.deltaT
  times, h = remove_amplitude_zero(times,h) # removes end portion of waveform which is zero
    
  return times, h


def EOBNRv2_LAL_modes(Mtot,q,dt,fmin,Dist,ell,emm):
  """ Simplified inferface to EOBNRv2 mode.

      INPUT
      =====
      Dist -- distance in megaparsecs.
      Mtot -- total mass in solar masses

      NOTE: modes are conjugated to agree with surrogate (FIXME) """
    
  from lalsimulation import SimInspiralChooseTDMode
  from lalsimulation import EOBNRv2

  Dist = Dist * 1e6 *PC_SI
  Mtot = Mtot * MSUN_SI
        
  M1,M2  = Mq_to_m1m2(Mtot,q)

  h      = SimInspiralChooseTDMode(phiRef=0.0,deltaT=dt,m1=M1,m2=M2,f_min=fmin,f_ref=0.0,r=Dist,\
                                   lambda1=0.0,lambda2=0.0,waveFlags=None,nonGRparams=None,amplitudeO=0,\
                                   phaseO=7,approximant=EOBNRv2,l=ell,m=emm)
    
  times = np.arange(np.size(h.data.data))*h.deltaT

  return times, np.conjugate(h.data.data) # <<-- NOTE (TODO) CONJUGATE


def fft_timeseries(h,df=0.0,verbose=False):
  """ Take an FFT of LAL datatype REAL8 or COMPLEX16 TimeSeries

      This routine was originally written by Evan Ochsner for lalsimwaves.py

     INPUT
     =====
     h -- time-series (e.g. strain hp*Fp + hc+Fc). h must be a REAL8TimeSeries
          which, for example, results from add hp and hc as returned 
          from  hp, hc = lalsim.SimInspiralChooseTDWaveform(...
     df     -- Setting df=0 means the TD waveform will just pads to the next power of 2.
               For df!=0, the TD waveform will be padded further, or it will print a
               warning if the waveform is so long that it requires a smaller df. """

  # to determine if h is real or complex timeseries
  from lal import REAL8TimeSeries, COMPLEX16TimeSeries
  assert(type(h)==REAL8TimeSeries or type(h)==COMPLEX16TimeSeries)

  from lal import ResizeREAL8TimeSeries, ResizeCOMPLEX16TimeSeries,\
                  CreateForwardREAL8FFTPlan, CreateForwardCOMPLEX16FFTPlan,\
                  REAL8TimeFreqFFT, COMPLEX16TimeFreqFFT,\
                  CreateCOMPLEX16FrequencySeries, HertzUnit

  sample = 1.0/h.deltaT # The sample rate

  print("sample rate = %f"%sample)

  # Find next power of 2 length #
  padlen = 1
  while padlen < h.data.length:
    padlen *= 2

  print("padlen (old) = %f"%padlen)
  # If df=0, use above padlen. If df != 0, then we... #
  if df != 0:
    # Check if padlen gives smaller df than requested
    # If so, print warning and use padlen
    if padlen > int(sample/df):
      print("Warning: waveform too long for requested df =", df)
      print("Instead using df =",  sample/padlen)  # padlen/sample
    # If requested df is finer freq. binning, then pad further
    if padlen < int(sample/df):
      padlen = int(sample/df)

  print("padlen (new) = %f"%padlen)

  # Pad h and create fft plan
  if (type(h) == REAL8TimeSeries):
    ResizeREAL8TimeSeries(h, 0, padlen)
    plan = CreateForwardREAL8FFTPlan(padlen, 0)
    hoff = CreateCOMPLEX16FrequencySeries(name="FD hplus",epoch=h.epoch,f0=0,\
                                          deltaF=1./(h.deltaT/h.data.length),\
                                          sampleUnits=HertzUnit,\
                                          length=int(padlen/2+1))
    REAL8TimeFreqFFT(hoff, h, plan)
  elif (type(h) == COMPLEX16TimeSeries):
    ResizeCOMPLEX16TimeSeries(h, 0, padlen)
    plan = CreateForwardCOMPLEX16FFTPlan(padlen,0)
    hoff = CreateCOMPLEX16FrequencySeries(name="FD hp+ihc",epoch=h.epoch,f0=0,\
                                        deltaF=1./(h.deltaT/h.data.length),\
                                        sampleUnits=HertzUnit,\
                                        length=padlen) #len OK?
    COMPLEX16TimeFreqFFT(hoff, h, plan)

  # deduce frequencies f such that (f,hoff(f))
  TDlen  = padlen
  FDlen  = int( hoff.data.length )
  siglen = TDlen*h.deltaT

  deltaF = 1/siglen
  frqs   = np.arange(FDlen) * deltaF
  if( type(h) == COMPLEX16TimeSeries):
    frqs = frqs - (sample/2.0)

  if verbose:
    print('TDlen = ',TDlen)
    print('TDlen/2 + 1 = ',TDlen/2+1)
    print('FDlen = ',FDlen)
    print('TDlen*dt = ',siglen)
    
  return frqs, hoff


def REAL8_TAPER_from_numpy(h,dt,sampleUnits,taper):
  """ From real timeseries h and deltaT dt, create a REAL8TimeSeries
      object and taper. 

      INPUT
      =====
      h           --- a real-valued numpy array representing the timeseries
      dt          --- time interval between samples
      sampleUnits --- lal.XXX units tag (e.g. lal.SecondUnit)
      taper       --- lalsim.XXX taper tag (e.g. lalsim.SIM_INSPIRAL_TAPER_STARTEND)"""

  from lal import CreateREAL8TimeSeries
  from lalsimulation import SimInspiralREAL8WaveTaper

  h_REAL8 = CreateREAL8TimeSeries("h(t)",epoch=0.0,f0=0.0,deltaT=dt,\
                                  sampleUnits=sampleUnits,length=len(h))
  h_REAL8.data.data[:] = h[:]

  SimInspiralREAL8WaveTaper(h_REAL8.data, taper)

  return h_REAL8


def COMPLEX16_TAPER_from_numpy(h,dt,sampleUnits,taper):
  """ From complex timeseries h and deltaT dt, create a Complex16TimeSeries
      object and taper. 

      INPUT
      =====
      h           --- a complex-valued numpy array representing the timeseries
      dt          --- time interval between samples
      sampleUnits --- lal.XXX units tag (e.g. lal.SecondUnit)
      taper       --- lalsim.XXX taper tag (e.g. lalsim.SIM_INSPIRAL_TAPER_STARTEND)"""

  from lal import CreateCOMPLEX16TimeSeries
  from lalsimulation import SIM_INSPIRAL_TAPER_NONE

  # LAL complex taper routine (~ SimInspiralCOMPLEX16WaveTaper) doesn't exist
  assert(taper==SIM_INSPIRAL_TAPER_NONE)

  h_COMPLEX16 = CreateCOMPLEX16TimeSeries("hp+ihc",epoch=0.0,f0=0,deltaT=dt,\
                                          sampleUnits=sampleUnits,length=len(h))
  h_COMPLEX16.data.data = h[:]

  return h_COMPLEX16


# TODO: should have set of routines which use numpy fft.fft (see Jonathan's code)
def fft_from_numpy(h,dt,sampleUnits,taper,df=0):
  """ From real or complex timeseries, h, and deltaT, dt, taper and
      perform an FFT using LAL code.

      INPUT
      =====
      h           --- a real or complex valued numpy array representing the timeseries
      dt          --- time interval between samples
      sampleUnits --- lal.XXX units tag (e.g. lal.SecondUnit)
      taper       --- lalsim.XXX taper tag (e.g. lalsim.SIM_INSPIRAL_TAPER_STARTEND).
                      To use a non-LAL taper, taper before passing h to this routine
                      and set taper = lalsim.SIM_INSPIRAL_TAPER_NONE

      OUTPUT
      ======
      hf --- FFT as a LAL timeseries (COMPLEX16FrequencySeries)"""


  # Taper before padding (could modify ringdown, which maybe good to do)
  if(h.dtype == np.double):
    h_REAL8  = REAL8_TAPER_from_numpy(h,dt,sampleUnits=sampleUnits,taper=taper)
    frqs, hf = fft_timeseries(h_REAL8,df,False)
  elif(h.dtype == np.complex):
    h_COMPLEX16 = COMPLEX16_TAPER_from_numpy(h,dt,sampleUnits,taper)
    frqs, hf = fft_timeseries(h_COMPLEX16,df,False)
  else:
    raise ValueError('dtype not supported')

  return frqs, hf


def convert_COMPLEX16FrequencySeries_to_2Sided_numpy(h,freqs):
  """ convert a lal data type COMPLEX16FrequencySeries, which stores a 1-sided 
      fft, to a 2-sided fft result a numpy array."""

  N = freqs.size
  hf_numpy   = np.array([h.data.data[N-i-1] if i<N else h.data.data[i-N+1] for i in np.arange((N-1)*2) ],dtype=complex)
  fs_2_sided = np.array([-freqs[N-i-1] if i<N else fs[i-N+1] for i in np.arange((N-1)*2) ])

  return fs_2_sided, hf_numpy


def make_psd_np_array(psd,freq_pts):
  """ Evaluate a canned LAL PSD at array of frequency points.

     INPUT
     =====
     psd      --- lalsimulation.SimNoisePSDXXX function 
     freq_pts --- array of frequency points """

  return np.array([ psd(f) for f in freq_pts],dtype=complex)


def whiten_numpy_waveform(psd,wv,freq_pts):
  """ Return whitened waveform h(f) / sqrt(S(f)) as a numpy array.

    INPUT
    =====
    psd      --- lalsimulation.SimNoisePSDXXX function
    wv       --- complex FFT waveform data (numpy array)
    freq_pts --- numpy array of frequency points """

  assert(len(wv) == len(freq_pts))
  return np.array([ wv[i] / np.sqrt( psd(freq_pts[i]) ) for i in np.arange(freq_pts.size)],dtype=complex,copy=True)


def whiten_COMPLEX16_waveform(psd,wv,freq_pts):
  """  Whiten a COMPLEX16 waveform  in place.

    INPUT
    =====
    psd --- lalsimulation.SimNoisePSDXXX function
    wv  --- complex FFT waveform data
    freq_pts --- numpy array of frequency points """

  wv_data = np.array(wv.data.data[:],dtype=complex)
  assert(len(wv_data) == len(freq_pts))
  wv.data.data[:]  = whiten_numpy_waveform(psd,wv_data,freq_pts)


def Fplus(theta, phi, psi):
  """
  "plus" antenna pattern function of sky location and polarization angle
  """
  return 0.5*(1. + np.cos(theta)**2)*np.cos(2.*phi)*np.cos(2.*psi) - np.cos(theta)*np.sin(2.*
phi)*np.sin(2.*psi)


def Fcross(theta, phi, psi):
  """
  "cross" antenna pattern function of sky location and polarization angle
  """
  return 0.5*(1. + np.cos(theta)**2)*np.cos(2.*phi)*np.sin(2.*psi) + np.cos(theta)*np.sin(2.*
phi)*np.cos(2.*psi)


def find_fft_deltaF(times):
  """Find an FFT-appropriate deltaF from array of times"""
  T = times.max() - times.min()
  powof2 = 1.0
  while( powof2 < T ):
      powof2 = powof2*2.0
      
  deltaF = 1.0/powof2
  print("Next power of 2 = %i"%powof2)
  print("DeltaF is %f"%deltaF)

  return deltaF



# spin-weighted spherical harmonics

#----------------------------------------------------------
#
# This module computes spin-weighted spherical harmonics.
#
# Released under the MIT License.
# (C) Christian Reisswig 2009-2011
#
#----------------------------------------------------------



from numpy import *



def fac(n):
   result = 1

   for i in range(2, n+1):
      result *= i

   return result





# coefficient function
def Cslm(s, l, m):
    return sqrt( l*l * (4.0*l*l - 1.0) / ( (l*l - m*m) * (l*l - s*s) ) )




# recursion function
def s_lambda_lm(s, l, m, x):

    Pm = pow(-0.5, m)

    if (m !=  s): Pm = Pm * pow(1.0+x, (m-s)*1.0/2)
    if (m != -s): Pm = Pm * pow(1.0-x, (m+s)*1.0/2)
   
    Pm = Pm * sqrt( fac(2*m + 1) * 1.0 / ( 4.0*pi * fac(m+s) * fac(m-s) ) )
   
    if (l == m):
        return Pm
   
    Pm1 = (x + s*1.0/(m+1) ) * Cslm(s, m+1, m) * Pm
   
    if (l == m+1):
        return Pm1
    else:
        for n in range (m+2, l+1):
            Pn = (x + s*m * 1.0 / ( n * (n-1.0) ) ) * Cslm(s, n, m) * Pm1 - Cslm(s, n, m) * 1.0 / Cslm(s, n-1, m) * Pm
            Pm = Pm1
            Pm1 = Pn  
        return Pn


def sYlm(ss, ll, mm, theta, phi):
   
    Pm = 1.0

    l = ll
    m = mm
    s = ss

    if (l < 0):
        return 0
    if (abs(m) > l or l < abs(s)):
        return 0

    if (abs(mm) < abs(ss)):
        s=mm
        m=ss
        if ((m+s) % 2):
            Pm  = -Pm

   
    if (m < 0):
        s=-s
        m=-m
        if ((m+s) % 2):
            Pm  = -Pm

    result = Pm * s_lambda_lm(s, l, m, cos(theta))

    return complex(result * cos(mm*phi), result * sin(mm*phi))





#import gwtools
from .gwtools import *
from .const import *
from .harmonics import *
try:
    from . import rotations
    from . import decompositions
    from . import fitfuncs
except ImportError:
    import warnings as _warnings
    _warnings.warn("Could not import rotations, decompositions, or fitfuncs. These are not needed by GWSurrogate.")




#!/usr/bin/env python

import numpy as np
import os
import unittest

if __package__ is "" or "None": # py2 and py3 compatible 
  print("setting __package__ to gwsurrogate.new so relative imports work")
  __package__="gwsurrogate.new"
from .saveH5Object import SimpleH5Object, H5ObjectList, H5ObjectDict, RESERVED_VALUE_STRINGS, RESERVED_KEY_STRINGS

TEST_FILE = 'test.h5' # Gets created and deleted

def _tear_down():
    if os.path.isfile(TEST_FILE):
        os.remove(TEST_FILE)

def _set_up():
    # Don't overwrite this in case it's actually needed by something else
    if os.path.isfile(TEST_FILE):
        raise Exception("{} already exists! Please move or remove it.")


class BaseTest(unittest.TestCase):

    def setUp(self):
        _set_up()

    def tearDown(self):
        _tear_down()


class DummyHolder(SimpleH5Object):

    def __init__(self, **kwargs):
        super(DummyHolder, self).__init__()

        for k, v in kwargs.items(): # inefficient in py2
            setattr(self, k, v)

class DummyHolderHolder(SimpleH5Object):

    def __init__(self, dummy_holders, **kwargs):
        sub_keys = ['dh_%s'%i for i in range(len(dummy_holders))]
        super(DummyHolderHolder, self).__init__(sub_keys=sub_keys)

        self.n_dummy_holders = len(dummy_holders)
        for i, dh in enumerate(dummy_holders):
            setattr(self, 'dh_%s'%(i), dh)

        for k, v in kwargs.items():  # inefficient in py2
            setattr(self, k, v)

    def h5_prepare_subs(self):
        for i in range(self.n_dummy_holders):
            setattr(self, 'dh_%s'%(i), DummyHolder())


class DataTester(BaseTest):

    def _test_items(self, items):
        c = DummyHolder(**items)
        c.save(TEST_FILE)
        c2 = DummyHolder(**{k: None for k in items.keys()})
        c2.load(TEST_FILE)
        for k in items.keys():
            v1 = getattr(c, k)
            v2 = getattr(c2, k)
            if type(v1) == dict:
                self.assertEqual(len(v1), len(v2))
                for k in v1.keys():
                    self.assertEqual(v1[k], v2[k])
            elif hasattr(v1, 'shape'):
                self.assertEqual(v1.shape, v2.shape)
            elif hasattr(v1, '__len__'):
                self.assertEqual(len(v1), len(v2))
                for i in range(len(v1)):
                    self.assertEqual(v1[i], v2[i])
            else:
                self.assertEqual(getattr(c, k), getattr(c2, k))

    def test_int(self):
        self._test_items({'zero': 0, 'one': 1, 'minus two': -2})

    def test_float(self):
        self._test_items({'zero': 0.0, 'one': 1.0, 'eps': 1.234e-15})

    def test_bool(self):
        self._test_items({'true': True, 'false': False})

    def test_string(self):
        self._test_items({'a': 'a', 'abc': 'abc', 'None': 'None'})

    def test_None(self):
        self._test_items({'None_type': None})

    def test_array(self):
        self._test_items({'number': np.array(1.2), 'empty1d': np.array([]),
                          '1d': np.arange(3), '2d': np.ones((3, 2, 4)),
                          'complex': np.exp(1.j*np.arange(3))})

    def test_list(self):
        self._test_items({'empty': [], 'assorted': ['string', 123, 0.52]})

    def test_tuple(self):
        self._test_items({'empty': (), 'assorted': ('string', 3, 0.5, None)})

    def test_dict(self):
        d = {'empty': {}, 'assorted': {'string': 'asdf', 'int': 1}}
        self._test_items(d)

class SubordinateTester(BaseTest):

    def test_subordinates(self):
        dh1 = DummyHolder(**{'a': 'abc', 'b': 0.3, 'c': -1})
        dh2 = DummyHolder(**{'d': 1234})
        dhh = DummyHolderHolder([dh1, dh2], **{'e': 321, 'f': 0.2})
        dhh.save(TEST_FILE)
        dhh2 = DummyHolderHolder([])
        dhh2.load(TEST_FILE)
        self.assertEqual(dhh.n_dummy_holders, dhh2.n_dummy_holders)
        dh1_2 = dhh2.dh_0
        dh2_2 = dhh2.dh_1
        self.assertEqual(dhh.e, dhh2.e)
        self.assertEqual(dhh.f, dhh2.f)
        for k in dh1.__dict__.keys():
            self.assertEqual(getattr(dh1, k), getattr(dh1_2, k))
        for k in dh2.__dict__.keys():
            self.assertEqual(getattr(dh2, k), getattr(dh2_2, k))
 
class ExceptionTester(BaseTest):

    def test_value_strings(self):
        for vstr in RESERVED_VALUE_STRINGS:
            with self.assertRaises(Exception):
                c = DummyHolder(asdf=vstr)
                c.save(TEST_FILE)
            _tear_down()

    def test_key_strings(self):
        for kstr in RESERVED_KEY_STRINGS:
            with self.assertRaises(Exception):
                kwargs = {kstr + 'asdf': 0.3}
                c = DummyHolder(**kwargs)
                c.save(TEST_FILE)

class H5ObjectListTester(BaseTest):

    def test_object_list(self):
        dhs = [DummyHolder(**{str(i): i*2}) for i in range(3)]
        h5ol = H5ObjectList(dhs)
        h5ol.save(TEST_FILE)
        h5ol2 = H5ObjectList([DummyHolder() for i in range(3)])
        h5ol2.load(TEST_FILE)
        for i in range(3):
            dh1 = dhs[i]
            dh2 = h5ol2[i]
            for k in dh1.__dict__.keys():
                self.assertEqual(getattr(dh1, k), getattr(dh2, k))

class H5ObjectDictTester(BaseTest):

    def test_object_list(self):
        dhs = {i: DummyHolder(**{str(i): i*2}) for i in range(1, 4)}
        h5ol = H5ObjectDict(dhs)
        h5ol.save(TEST_FILE)
        h5ol2 = H5ObjectDict({i: DummyHolder() for i in range(1, 4)})
        h5ol2.load(TEST_FILE)
        for i in range(1, 4):
            dh1 = dhs[i]
            dh2 = h5ol2[i]
            for k in dh1.__dict__.keys():
                self.assertEqual(getattr(dh1, k), getattr(dh2, k))




""" Gravitational Wave Surrogate classes for text and hdf5 files"""

from __future__ import division # for python 2

__copyright__ = "Copyright (C) 2014 Scott Field and Chad Galley"
__email__     = "sfield@astro.cornell.edu, crgalley@tapir.caltech.edu"
__status__    = "testing"
__author__    = "Scott Field, Chad Galley"

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

# adding "_" prefix to potentially unfamiliar module names
# so they won't show up in gws' tab completion
import numpy as np
from scipy.interpolate import splrep as _splrep
from scipy.interpolate import splev as _splev
from .gwtools.harmonics import sYlm as _sYlm
from .gwtools import plot_pretty as _plot_pretty
from .gwtools import gwtools as _gwtools # from the package gwtools, import the module gwtools (gwtools.py)....
from .parametric_funcs import function_dict as my_funcs
from .surrogateIO import H5Surrogate as _H5Surrogate
from .surrogateIO import TextSurrogateRead as _TextSurrogateRead
from .surrogateIO import TextSurrogateWrite as _TextSurrogateWrite

try:
  import matplotlib.pyplot as plt
except:
  print("Cannot load matplotlib.")

try:
  import h5py
  h5py_enabled = True
except ImportError:
  h5py_enabled = False


# needed to search for single mode surrogate directories 
def _list_folders(path,prefix):
  '''returns all folders which begin with some prefix'''
  import os as os
  for f in os.listdir(path):
    if f.startswith(prefix):
      yield f

# handy helper to save waveforms 
def write_waveform(t, hp, hc, filename='output',ext='bin'):
  """write waveform to text or numpy binary file"""

  if( ext == 'txt'):
    np.savetxt(filename, [t, hp, hc])
  elif( ext == 'bin'):
    np.save(filename, [t, hp, hc])
  else:
    raise ValueError('not a valid file extension')


##############################################
class ExportSurrogate(_H5Surrogate, _TextSurrogateWrite):
	"""Export single-mode surrogate"""
	
	#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
	def __init__(self, path):
		
		# export HDF5 or Text surrogate data depending on input file extension
		ext = path.split('.')[-1]
		if ext == 'hdf5' or ext == 'h5':
			_H5Surrogate.__init__(self, file=path, mode='w')
		else:
			raise ValueError('use TextSurrogateWrite instead')


##############################################
class EvaluateSingleModeSurrogate(_H5Surrogate, _TextSurrogateRead):
  """Evaluate single-mode surrogate in terms of the waveforms' amplitude and phase"""


  #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  def __init__(self, path, deg=3, subdir='', closeQ=True):

    # Load HDF5 or Text surrogate data depending on input file extension
    if type(path) == h5py._hl.files.File:
      ext = 'h5'
    else:
      ext = path.split('.')[-1]
    if ext == 'hdf5' or ext == 'h5':
      _H5Surrogate.__init__(self, file=path, mode='r', subdir=subdir, closeQ=closeQ)
    else:
      _TextSurrogateRead.__init__(self, path)
    
    # Interpolate columns of the empirical interpolant operator, B, using cubic spline
    if self.surrogate_mode_type  == 'waveform_basis':
      self.reB_spline_params = [_splrep(self.times, self.B[:,jj].real, k=deg) for jj in range(self.B.shape[1])]
      self.imB_spline_params = [_splrep(self.times, self.B[:,jj].imag, k=deg) for jj in range(self.B.shape[1])]
    elif self.surrogate_mode_type  == 'amp_phase_basis':
      self.B1_spline_params = [_splrep(self.times, self.B_1[:,jj], k=deg) for jj in range(self.B_1.shape[1])]
      self.B2_spline_params = [_splrep(self.times, self.B_2[:,jj], k=deg) for jj in range(self.B_2.shape[1])]
    else:
      raise ValueError('invalid surrogate type')

    pass

  #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  def __call__(self, q, M=None, dist=None, phi_ref=None,\
                     f_low=None, samples=None,samples_units='dimensionless',\
                     singlemode_call=True):
    """Return single mode surrogate evaluation for...

       Input
       =====
       q               --- binary parameter values EXCLUDING total mass M.
                           In 1D, mass ratio (dimensionless) must be supplied.
                           In nD, the surrogate's internal parameterization is assumed.
       M               --- total mass (solar masses) 
       dist            --- distance to binary system (megaparsecs)
       phi_ref         --- mode's phase phi(t), as h=A*exp(i*phi) at peak amplitude
       f_low           --- instantaneous initial frequency, will check if flow_surrogate < f_low
       samples         --- array of times at which surrogate is to be evaluated
       samples_units   --- units (mks or dimensionless) of input array samples
       singlemode_call --- Never set this by hand, will be False if this routine is called by the multimode evaluator


       More information
       ================
       This routine evaluates gravitational wave complex polarization modes h_{ell m}
       defined on a sphere whose origin is the binary's center of mass. 

       Dimensionless surrogates rh/M are evaluated by calling _h_sur. 
       Physical surrogates are generated by applying additional operations/scalings.

       If M and dist are provided, a physical surrogate will be returned in mks units.

       An array of times can be passed along with its units. """

    # surrogate evaluations assumed dimensionless, physical modes are found from scalings 
    if self.surrogate_units != 'dimensionless':
      raise ValueError('surrogate units is not supported')

    if (samples_units != 'dimensionless') and (samples_units != 'mks'):
      raise ValueError('samples_units is not supported')

    ### if (M,distance) provided, a physical mode in mks units is returned ###
    if( M is not None and dist is not None):
      amp0    = ((M * _gwtools.MSUN_SI ) / (dist * _gwtools.PC_SI )) * ( _gwtools.G / np.power(_gwtools.c,2.0) )
      t_scale = _gwtools.Msuninsec * M
    else:
      amp0    = 1.0
      t_scale = 1.0

    ### evaluation times t: input times or times at which surrogate was built ###
    if (samples is not None):
      t = samples
    else:
      t = self.time()

    ### if input times are dimensionless, convert to MKS if a physical surrogate is requested ###
    if samples_units == 'dimensionless':
      t = t_scale * t

    # because samples is passed to _h_sur, it must be dimensionless form t/M
    if samples is not None and samples_units == 'mks':
      samples = samples / t_scale

    # convert from input to internal surrogate parameter values, and check within training region #
    x = self.get_surr_params_safe(q)

    ### Evaluate dimensionless single mode surrogates ###
    hp, hc = self._h_sur(x, samples=samples)


    ### adjust mode's phase by an overall constant ###
    if (phi_ref is not None):
      h  = self.adjust_merger_phase(hp + 1.0j*hc,phi_ref)
      hp = h.real
      hc = h.imag

    ### Restore amplitude scalings ###
    hp     = amp0 * hp
    hc     = amp0 * hc

    ### check that surrogate's starting frequency is below f_low, otherwise throw a warning ###
    if f_low is not None:
      self.find_instant_freq(hp, hc, t, f_low)

    return t, hp, hc


  #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  def find_instant_freq(self, hp, hc, t, f_low = None):
    """instantaneous frequency at t_start for 

          h = A(t) exp(2 * pi * i * f(t) * t), 

       where \partial_t A ~ \partial_t f ~ 0. If f_low passed will check its been achieved."""

    f_instant = _gwtools.find_instant_freq(hp, hc, t)

    # TODO: this is a hack to account for inconsistent conventions!
    f_instant = np.abs(f_instant)

    if f_low is None:
      return f_instant
    else:
      if f_instant > f_low:
        raise Warning("starting frequency is "+str(f_instant))
      else:
        pass


  #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  def amp_phase(self,h):
    """Get amplitude and phase of waveform, h = A*exp(i*phi)"""
    return _gwtools.amp_phase(h)


  #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  def phi_merger(self,h):
    """Phase of mode at amplitude's discrete peak. h = A*exp(i*phi)."""

    amp, phase = self.amp_phase(h)
    argmax_amp = np.argmax(amp)

    return phase[argmax_amp]


  #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  def adjust_merger_phase(self,h,phiref):
    """Modify GW mode's phase such that at time of amplitude peak, t_peak, we have phase(t_peak) = phiref"""

    phimerger = self.phi_merger(h)
    phiadj    = phiref - phimerger

    return _gwtools.modify_phase(h,phiadj)


  #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  def timer(self,M_eval=None,dist_eval=None,phi_ref=None,f_low=None,samples=None):
    """average time to evaluate surrogate waveforms. """

    qmin, qmax = self.fit_interval
    ran = np.random.uniform(qmin, qmax, 1000)

    import time
    tic = time.time()
    if M_eval is None:
      for i in ran:
        hp, hc = self._h_sur(i)
    else:
      for i in ran:
        t, hp, hc = self.__call__(i,M_eval,dist_eval,phi_ref,f_low,samples)

    toc = time.time()
    print('Timing results (results quoted in seconds)...')
    print('Total time to generate 1000 waveforms = ',toc-tic)
    print('Average time to generate a single waveform = ', (toc-tic)/1000.0)
    pass

	
  #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  def time(self, units=None,M=None,dt=None):
    """Return the set of time samples at which the surrogate is defined.
       If dt is supplied, return the requested uniform grid.

      INPUT (all optional)
      =====
      units --- None:        time in geometric units, G=c=1
                'mks'        time in seconds
                'solarmass': time in solar masses
      M     --- Mass (in units of solar masses).
      dt    --- delta T

      OUTPUT
      ======
      1) units = M = None:   Return time samples at which the surrogate as built for.
      2) units != None, M=:  Times after we convert from surrogate's self.t_units to units.
                             If units = 'mks' and self.t_units='TOverMtot' then M must
                             be supplied to carry out conversion.
      3) dt != None:         Return time grid as np.arange(t[0],t[-1],dt)"""


    if units is None:
      t = self.times
    elif (units == 'mks') and (self.t_units == 'TOverMtot'):
      assert(M!=None)
      t = (_gwtools.Msuninsec*M) * self.times
    else:
      raise ValueError('Cannot compute times')

    if dt is None:
      return t
    else:
      return np.arange(t[0],t[-1],dt)

  #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  def basis(self, i, flavor='waveform'):
    """compute the ith cardinal, orthogonal, or waveform basis."""

    if self.surrogate_mode_type  == 'waveform_basis':

      if flavor == 'cardinal':
        basis = self.B[:,i]
      elif flavor == 'orthogonal':
        basis = np.dot(self.B,self.V)[:,i]
      elif flavor == 'waveform':
        E = np.dot(self.B,self.V)
        basis = np.dot(E,self.R)[:,i]
      else:
        raise ValueError("Not a valid basis type")

    elif self.surrogate_mode_type  == 'amp_phase_basis':
        raise ValueError("Not coded yet")

    return basis

  # TODO: basis resampling should be unified.
  #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  # TODO: ext should be passed from __call__
  def resample_B(self, samples, ext=1):
    """resample the empirical interpolant operator, B, at the input samples"""

    evaluations = np.array([_splev(samples, self.reB_spline_params[jj],ext=ext)  \
             + 1j*_splev(samples, self.imB_spline_params[jj],ext=ext) for jj in range(self.B.shape[1])]).T

    # allow for extrapolation if very close to surrogate's temporal interval
    t0 = self.times[0]
    if (np.abs(samples[0] - t0) < t0 * 1.e-12) or (t0==0 and np.abs(samples[0] - t0) <1.e-12):
      evaluations[0] = np.array([_splev(samples[0], self.reB_spline_params[jj],)  \
             + 1j*_splev(samples[0], self.imB_spline_params[jj]) for jj in range(self.B.shape[1])]).T
    
    return evaluations

  #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  # TODO: ext should be passed from __call__
  def resample_B_1(self, samples, ext=1):
    """resample the B_1 basis at the input samples"""

    evaluations = np.array([_splev(samples, self.B1_spline_params[jj],ext=1) for jj in range(self.B_1.shape[1])]).T

    # allow for extrapolation if very close to surrogate's temporal interval
    if np.abs(samples[0] - self.times[0])/self.times[0] < 1.e-12:
      evaluations[0] = np.array([_splev(samples[0], self.B1_spline_params[jj],ext=1) for jj in range(self.B_1.shape[1])]).T
    
    return evaluations

  #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  # TODO: ext should be passed from __call__
  def resample_B_2(self, samples, ext=1):
    """resample the B_2 basis at the input samples"""

    evaluations = np.array([_splev(samples, self.B2_spline_params[jj],ext=1) for jj in range(self.B_2.shape[1])]).T

    # allow for extrapolation if very close to surrogate's temporal interval
    if np.abs(samples[0] - self.times[0])/self.times[0] < 1.e-12:
      evaluations[0] = np.array([_splev(samples[0], self.B2_spline_params[jj],ext=1) for jj in range(self.B_2.shape[1])]).T
    
    return evaluations


  #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  def plot_rb(self, i, showQ=True):
    """plot the ith reduced basis waveform"""

    # Compute surrogate approximation of RB waveform
    basis = self.basis(i)
    fig   = _plot_pretty(self.times,[basis.real,basis.imag])

    if showQ:
      plt.show()
    
    # Return figure method to allow for saving plot with fig.savefig
    return fig


  #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  def plot_sur(self, q_eval, timeM=False, htype='hphc', flavor='linear', color='k', linestyle=['-', '--'], \
                label=['$h_+(t)$', '$h_-(t)$'], legendQ=False, showQ=True):
    """plot surrogate evaluated at mass ratio q_eval"""

    t, hp, hc = self.__call__(q_eval)
    h = hp + 1j*hc
    
    y = {
      'hphc': [hp, hc],
      'hp': hp,
      'hc': hc,
      'AmpPhase': [np.abs(h), _gwtools.phase(h)],
      'Amp': np.abs(h),
      'Phase': _gwtools.phase(h),
      }
    
    if self.t_units == 'TOverMtot':
      xlab = 'Time, $t/M$'
    else:
      xlab = 'Time, $t$ (sec)'

    # Plot surrogate waveform
    fig = _plot_pretty(t, y[htype], flavor=flavor, color=color, linestyle=linestyle, \
                label=label, legendQ=legendQ, showQ=False)
    plt.xlabel(xlab)
    plt.ylabel('Surrogate waveform')
    
    if showQ:
      plt.show()
        
    # Return figure method to allow for saving plot with fig.savefig
    return fig
  
  
  #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  def plot_eim_data(self, inode=None, htype='Amp', nuQ=False, fignum=1, showQ=True):
    """Plot empirical interpolation data used for performing fits in parameter"""
    
    fig = plt.figure(fignum)
    ax1 = fig.add_subplot(111)
    
    y = {
      'Amp': self.eim_amp,
      'Phase': self.eim_phase,
      }
    
    if nuQ:
      nu = _gwtools.q_to_nu(self.greedy_points)
      
      if inode is None:
        [plt.plot(nu, ee, 'ko') for ee in y[htype]]
      else:
        plt.plot(nu, y[htype][inode], 'ko')
      
      plt.xlabel('Symmetric mass ratio, $\\nu$')
    
    else:
      
      if inode is None:
        [plt.plot(self.greedy_points, ee, 'ko') for ee in y[htype]]
      else:
        plt.plot(self.greedy_points, y[htype][inode], 'ko')
      
      plt.xlabel('Mass ratio, $q$')
    
    if showQ:
      plt.show()
    
    return fig
  
  
  #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  def plot_eim_fits(self, inode=None, htype='Amp', nuQ=False, fignum=1, num=200, showQ=True):
    """Plot empirical interpolation data and fits"""
    
    fig = plt.figure(fignum)
    ax1 = fig.add_subplot(111)
    
    fitfn = {
      'Amp': self.amp_fit_func,
      'Phase': self.phase_fit_func,
      }
    
    coeffs = {
      'Amp': self.fitparams_amp,
      'Phase': self.fitparams_phase,
      }
    
    # Plot EIM data points
    self.plot_eim_data(inode=inode, htype=htype, nuQ=nuQ, fignum=fignum, showQ=False)
    
    qs = np.linspace(self.fit_min, self.fit_max, num)
    nus = _gwtools.q_to_nu(qs)
    
    # Plot fits to EIM data points
    if nuQ:
      if inode is None:
        [plt.plot(nus, fitfn[htype](cc, qs), 'k-') for cc in coeffs[htype]]
      else:
        plt.plot(nus, fitfn[htype](coeffs[htype][inode], qs), 'k-')  
    
    else:
      if inode is None:
        [plt.plot(qs, fitfn[htype](cc, qs), 'k-') for cc in coeffs[htype]]
      else:
        plt.plot(qs, fitfn[htype](coeffs[htype][inode], qs), 'k-')
      
    if showQ:
      plt.show()
    
    return fig
  

  #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  # TODO: strong_checking should be kwargs
  # TODO: only check once in multimode surrogates
  def check_training_interval(self, x, strong_checking=True):
    """Check if parameter value x is within the training interval."""

    x_min, x_max = self.fit_interval

    if(np.any(x < x_min) or np.any(x > x_max)):
      if strong_checking:
        raise ValueError('Surrogate not trained at requested parameter value')
      else:
        print("Warning: Surrogate not trained at requested parameter value")
        Warning("Surrogate not trained at requested parameter value")


  #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  def get_surr_params_safe(self,x):
    """ Compute the surrogate's *internal* parameter values from the input ones, x,
        passed to __call__ with safe bounds checking.

        The function get_surr_params used in the conversion is set in SurrogateIO
        as specified by the surrogate's data value corresponding to the key PARAMETERIZATION.
        Therefore, SurrogateIO must be aware of what x is expected to be. 

          Example: The user may pass mass ratio q=x to __call__, but the
                   symmetric mass ratio x_internal = q / (1+q)^2 might parameterize the surrogate

        After the parameter change of coordinates is done, check its within the surrogate's
        training region. Training regions is assumed to be an n-dim rectangle.

        x is assumed to NOT have total mass M as a parameter. ``Bare" surrogates are always dimensionless."""

    x_internal = self.get_surr_params(x)

    # TODO: this will (redundently) check for each mode. Multimode surrogate should directly check it
    self.check_training_interval(x_internal)

    return x_internal

  #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  def norm_eval(self, x):
    """Evaluate norm fit at parameter value x.

       Wrapper for norm evaluations called from outside of the class"""

    self.check_training_interval(x, strong_checking=True)
    x_0 = self._affine_mapper(x) # _norm_eval won't do its own affine mapping
    return self._norm_eval(x_0)

  #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  def eim_coeffs(self, x, surrogate_mode_type):
    """Evaluate EIM coefficients at parameter value x.

       Wrapper for safe calls from outside of the class"""

    self.check_training_interval(x, strong_checking=True)
    return self._eim_coeffs(x, surrogate_mode_type)

  #### below here are "private" member functions ###
  # These routine's evaluate a "bare" surrogate, and should only be called
  # by the __call__ method 
  #
  # These routine's use x as the parameter, which could be mass ratio,
  # symmetric mass ratio, or something else. Parameterization info should
  # be supplied by surrogate's parameterization tag.

  #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  def _affine_mapper(self, x):
    """map parameter value x to the standard interval [-1,1] if necessary."""

    x_min, x_max = self.fit_interval

    if self.affine_map == 'minus1_to_1':
      x_0 = 2.*(x - x_min)/(x_max - x_min) - 1.;
    elif self.affine_map == 'zero_to_1':
      x_0 = (x - x_min)/(x_max - x_min);
    elif self.affine_map == 'none':
      x_0 = x
    else:
      raise ValueError('unknown affine map')
    return x_0


  #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  def _norm_eval(self, x_0):
    """Evaluate norm fit at x_0, where x_0 is the mapped parameter value.

       WARNING: this function should NEVER be called from outside the class."""

    if not self.norms:
      return 1.
    else:
      return np.array([ self.norm_fit_func(self.fitparams_norm, x_0) ])


  #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  def _amp_eval(self, x_0):
    """Evaluate set of amplitude fits at x_0, where x_0 is the mapped parameter value.

       WARNING: this function should NEVER be called from outside the class."""

    if self.fit_type_amp == 'fast_spline_real':
      return self.amp_fit_func(self.fitparams_amp, x_0)
    else:
      return np.array([ self.amp_fit_func(self.fitparams_amp[jj,:], x_0) for jj in range(self.fitparams_amp.shape[0]) ])


  #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  def _phase_eval(self, x_0):
    """Evaluate set of phase fit at x_0, where x_0 is the mapped parameter value.

       WARNING: this function should NEVER be called from outside the class."""

    if self.fit_type_phase == 'fast_spline_imag':
      return self.phase_fit_func(self.fitparams_phase, x_0)
    else:
      return np.array([ self.phase_fit_func(self.fitparams_phase[jj,:], x_0) for jj in range(self.fitparams_phase.shape[0]) ])


  #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  def _eim_coeffs(self, x, surrogate_mode_type):
    """Evaluate EIM coefficients at parameter value x. x could be mass ratio, symmetric
       mass ratio or something else -- it depends on the surrogate's parameterization. 

       see __call__ for the parameterization and _h_sur for how these 
       coefficients are used. 

       If called from outside the class, check_training_interval should be used
       to determine whether x is in the training interval.

       WARNING: this function should NEVER be called from outside the class."""


    ### x to the standard interval on which the fits were performed ###
    x_0 = self._affine_mapper(x)

    ### Evaluate amp/phase/norm fits ###
    amp_eval   = self._amp_eval(x_0)
    phase_eval = self._phase_eval(x_0)
    nrm_eval   = self._norm_eval(x_0)

    if self.surrogate_mode_type  == 'waveform_basis':
      if self.fit_type_amp == 'fast_spline_real':
        h_EIM = nrm_eval * (amp_eval + 1j*phase_eval)
      else:
        h_EIM = nrm_eval*amp_eval*np.exp(1j*phase_eval) # dim_RB-vector fit evaluation of h
      return h_EIM
    elif self.surrogate_mode_type  == 'amp_phase_basis':
      if self.fit_type_amp == 'fast_spline_real':
        raise ValueError("invalid combination")
      return amp_eval, phase_eval, nrm_eval
    else: 
      raise ValueError('invalid surrogate type')

  #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  def _h_sur(self, x, samples=None):
    """Evaluate surrogate at parameter value x. x could be mass ratio, symmetric
       mass ratio or something else -- it depends on the surrogate's parameterization. 

       Returns dimensionless rh/M waveforms in units of t/M.

       This should ONLY be called by the __call__ method which accounts for 
       different parameterization choices. """


    if self.surrogate_mode_type  == 'waveform_basis':

      h_EIM = self._eim_coeffs(x, 'waveform_basis')

      if samples is None:
        surrogate = np.dot(self.B, h_EIM)
      else:
        surrogate = np.dot(self.resample_B(samples), h_EIM)

      #surrogate = nrm_eval * surrogate

    elif self.surrogate_mode_type  == 'amp_phase_basis':

      amp_eval, phase_eval, nrm_eval = self._eim_coeffs(x, 'amp_phase_basis')

      if samples is None:
        sur_A = np.dot(self.B_1, amp_eval)
        sur_P = np.dot(self.B_2, phase_eval)
      else:
        #sur_A = np.dot(np.array([_splev(samples, self.B1_spline_params[jj],ext=1) for jj in range(self.B_1.shape[1])]).T, amp_eval)
        #sur_P = np.dot(np.array([_splev(samples, self.B2_spline_params[jj],ext=1) for jj in range(self.B_2.shape[1])]).T, phase_eval)
        sur_A = np.dot(self.resample_B_1(samples), amp_eval)
        sur_P = np.dot(self.resample_B_2(samples), phase_eval)

      surrogate = nrm_eval * sur_A * np.exp(1j*sur_P)


    else:
      raise ValueError('invalid surrogate type')


    hp = surrogate.real
    #hp = hp.reshape([self.time_samples,])
    hc = surrogate.imag
    #hc = hc.reshape([self.time_samples,])

    return hp, hc


def CreateManyEvaluateSingleModeSurrogates(path, deg, ell_m, excluded, enforce_orbital_plane_symmetry):
  """For each surrogate mode an EvaluateSingleModeSurrogate class
     is created.

     INPUT
     =====
     path: the path to the surrogate
     deg: the degree of the splines representing the basis (default 3, cubic)
     ell_m: A list of (ell, m) modes to load, for example [(2,2),(3,3)].
            None (default) loads all modes.
     excluded: A list of (ell, m) modes to skip loading.
        The default ('DEFAULT') excludes any modes with an 'EXCLUDED' dataset.
        Use [] or None to load these modes as well.
     enforce_orbital_plane_symmetry: If set to True an exception is raised if the 
        surrogate data contains negative modes. This can be used to gaurd against
        mixing spin-aligned and precessing surrogates...which have different
        evaluation patterns for m<0.

     Returns single_mode_dict. Keys are (ell, m) mode and value is an
     instance of EvaluateSingleModeSurrogate."""


  if excluded is None:
    excluded = []

  ### fill up dictionary with single mode surrogate class ###
  single_mode_dict = dict()

  # Load HDF5 or Text surrogate data depending on input file extension
  if type(path) == h5py._hl.files.File:
    ext = 'h5'
    filemode = path.mode
  else:
    ext = path.split('.')[-1]
    filemode = 'r'

  # path, excluded 
  if ext == 'hdf5' or ext == 'h5':
      
    if filemode not in ['r+', 'w']:
      fp = h5py.File(path, filemode)

      ### compile list of excluded modes ###
      if type(excluded) == list:
        exc_modes = excluded
      elif excluded == 'DEFAULT':
        exc_modes = []
      else:
        raise ValueError('Invalid excluded option: %s'%excluded)
      for kk, vv in fp.items():  # inefficient on Py2
        if 'EXCLUDED' in vv:
          splitkk = kk.split('_')
          if splitkk[0][0] == 'l' and splitkk[1][0] == 'm':
            ell = int(splitkk[0][1])
            emm = int(splitkk[1][1:])
            if excluded == 'DEFAULT':
              exc_modes.append((ell,emm))
            elif not (ell, emm) in exc_modes:
              print("Warning: Including mode (%d,%d) which is excluded by default"%(ell, emm))
       ### compile list of available modes ###
      if ell_m is None:
        mode_keys = []
        for kk in fp.keys(): # Py2 list, Py3 iterator
          splitkk = kk.split('_')
          if splitkk[0][0] == 'l' and splitkk[1][0] == 'm':
            ell = int(splitkk[0][1])
            emm = int(splitkk[1][1:])
            if not (ell, emm) in exc_modes:
              mode_keys.append((ell,emm))
      else:
        mode_keys = []
        for i, mode in enumerate(ell_m):
          if mode in exc_modes:
            print("WARNING: Mode (%d,%d) is both included and excluded! Excluding it."%mode) 
          else:
            mode_keys.append(mode)

      # If we are using orbital symmetry, make sure we aren't loading any negative m modes
      if enforce_orbital_plane_symmetry:
        for ell, emm in mode_keys:
          if emm < 0:
            raise Exception("When using enforce_orbital_plane_symmetry, do not load negative m modes!")

       ### load the single mode surrogates ###
      for mode_key in mode_keys:
        assert(mode_keys.count(mode_key)==1)
        mode_key_str = 'l'+str(mode_key[0])+'_m'+str(mode_key[1])
        print("loading surrogate mode... " + mode_key_str)
        single_mode_dict[mode_key] = \
          EvaluateSingleModeSurrogate(fp,subdir=mode_key_str+'/',closeQ=False)
      fp.close()

  else:
    ### compile list of available modes ###
    # assumes (i) single mode folder format l#_m#_ 
    #         (ii) ell<=9, m>=0
    import os
    for single_mode in _list_folders(path,'l'):
      ell = int(single_mode[1])
      emm = int(single_mode[4])
      mode_key = (ell,emm)
      if (ell_m is None) or (mode_key in ell_m):
        if ((type(excluded) == list and not mode_key in excluded) or
            (excluded == 'DEFAULT' and not
             os.path.isfile(path+single_mode+'/EXCLUDED.txt'))):
          assert(mode_key not in single_mode_dict)
          if os.path.isfile(path+single_mode+'/EXCLUDED.txt'):
            print("Warning: Including mode (%d,%d) which is excluded by default"%(ell, emm))
          if enforce_orbital_plane_symmetry and emm < 0:
            raise Exception("When using enforce_orbital_plane_symmetry, do not load negative m modes!")

          print("loading surrogate mode... "+single_mode[0:5])
          single_mode_dict[mode_key] = \
            EvaluateSingleModeSurrogate(path+single_mode+'/')
    ### check all requested modes have been loaded ###
    if ell_m is not None:
      for tmp in ell_m:
        try:
          single_mode_dict[tmp]
        except KeyError:
          print('Could not find mode '+str(tmp))
 
  return single_mode_dict


##############################################
class EvaluateSurrogate():
  """Evaluate multi-mode surrogates"""

  #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  def __init__(self, path, deg=3, ell_m=None, excluded='DEFAULT', use_orbital_plane_symmetry=True):
    """Loads a surrogate.

    path: the path to the surrogate
    deg: the degree of the splines representing the basis (default 3, cubic). 
        Unless there is good reason to use deg !=3 one should not change this.
        Some surrogates (e.g. 4d2s) are validated with this in mind.
    ell_m: A list of (ell, m) modes to load, for example [(2,2),(3,3)].
        None (default) loads all modes.
    excluded: A list of (ell, m) modes to skip loading.
        The default ('DEFAULT') excludes any modes with an 'EXCLUDED' dataset.
        Use [] or None to load these modes as well.
    use_orbital_plane_symmetry: If set to true (i) CreateManyEvaluateSingleModeSurrogates
        will explictly check that m<0 do not exist in the data file and (ii) m<0 modes
        are inferred from m>0 modes. If set to false no symmetry is assumed -- typical
        of precessing models. When False, fake_neg_modes must be false."""

    self.single_mode_dict = \
      CreateManyEvaluateSingleModeSurrogates(path, deg, ell_m, excluded, use_orbital_plane_symmetry)

    self.use_orbital_plane_symmetry = use_orbital_plane_symmetry

    ### Load/deduce multi-mode surrogate properties ###
    #if filemode not in ['r+', 'w']:      
    if len(self.single_mode_dict) == 0:
      raise IOError('Modes not found. Mode subdirectories begins with l#_m#_')

    ### Check single mode temporal grids are collocated ###
    grid_shape = self.single_mode_dict[list(self.single_mode_dict.keys())[0]].time().shape
    for key in list(self.single_mode_dict.keys()):
      tmp_shape = self.single_mode_dict[list(self.single_mode_dict.keys())[0]].time().shape
      if(grid_shape != tmp_shape):
        raise ValueError('inconsistent single mode temporal grids')
      
    self.time_all_modes = self.single_mode_dict[list(self.single_mode_dict.keys())[0]].time

      # TODO: other common data to merge: surrogate parameter intervals
      # Bad idea? if modes use different parameterization -- better to let modes handle this?

  #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  def __call__(self, q, M=None, dist=None, theta=None,phi=None,\
                     z_rot=None, f_low=None, samples=None,\
                     samples_units='dimensionless',\
                     ell=None, m=None, mode_sum=True,fake_neg_modes=True):
    """Return surrogate evaluation for...

      INPUT
      =====
      q              --- binary parameter values EXCLUDING total mass M.
                         In 1D, mass ratio (dimensionless) must be supplied.
                         In nD, the surrogate's internal parameterization is assumed.
      M              --- total mass (solar masses) 
      dist           --- distance to binary system (megaparsecs)
      theta/phi      --- evaluate hp and hc modes at this location on sphere
      z_rot          --- physical rotation about angular momentum (z-)axis (radians)
      flow           --- instantaneous initial frequency, will check if flow_surrogate < flow mode-by-mode
      samples        --- array of times at which surrogate is to be evaluated
      samples_units  --- units ('mks' or 'dimensionless') of input array samples
      ell            --- list or array of N ell modes to evaluate for (if none, all modes are returned)
      m              --- for each ell, supply a matching m value 
      mode_sum       --- if true, all modes are summed, if false all modes are returned in an array
      fake_neg_modes --- if true, include m<0 modes deduced from m>0 mode. all m in [ell,m] input should be non-negative

      NOTE: if only requesting one mode, this should be ell=[2],m=[2]

       Note about Angles
       =================
       For circular orbits, the binary's orbital angular momentum is taken to
       be the z-axis. Theta and phi is location on the sphere relative to this 
       coordinate system. """


    if (not self.use_orbital_plane_symmetry) and fake_neg_modes:
      raise ValueError("if use_orbital_plane_symmetry is not assumed, it is not possible to fake m<0 modes")

    ### deduce single mode dictionary keys from ell, m and fake_neg_modes input ###
    modes_to_evaluate = self.generate_mode_eval_list(ell,m,fake_neg_modes)

    if mode_sum and (theta is None and phi is None) and len(modes_to_evaluate)!=1:
      raise ValueError('Trying to sum modes without theta and phi is a strange idea')

    ### if mode_sum false, return modes in a sensible way ###
    if not mode_sum:
      modes_to_evaluate = self.sort_mode_list(modes_to_evaluate)

    # Modes actually modeled by the surrogate. We will fake negative m
    # modes later if needed. 
    modeled_modes = self.all_model_modes(False)

    ### allocate arrays for multimode polarizations ###
    if mode_sum:
      hp_full, hc_full = self._allocate_output_array(samples,1,mode_sum)
    else:
      hp_full, hc_full = self._allocate_output_array(samples,len(modes_to_evaluate),mode_sum)

    ### loop over all evaluation modes ###
    # TODO: internal workings are simplified if h used instead of (hc,hp)
    ii = 0
    for ell,m in modes_to_evaluate:

      ### if the mode is modelled, evaluate it. Otherwise its zero ###
      is_modeled = (ell,m) in modeled_modes
      neg_modeled = (ell,-m) in modeled_modes
      if is_modeled or (neg_modeled and fake_neg_modes):

        if is_modeled:
          t_mode, hp_mode, hc_mode = self.evaluate_single_mode(q,M,dist,f_low,samples,samples_units,ell,m)
        else: # then we must have neg_modeled=True and fake_neg_modes=True
          t_mode, hp_mode, hc_mode = self.evaluate_single_mode_by_symmetry(q,M,dist,f_low,samples,samples_units,ell,m)

        if z_rot is not None:
          h_tmp   = _gwtools.modify_phase(hp_mode+1.0j*hc_mode,z_rot*m)
          hp_mode = h_tmp.real
          hc_mode = h_tmp.imag

        # TODO: should be faster. integrate this later on
        #if fake_neg_modes and m != 0:
        #  hp_mode_mm, hc_mode_mm = self._generate_minus_m_mode(hp_mode,hc_mode,ell,m)
        #  hp_mode_mm, hc_mode_mm = self.evaluate_on_sphere(ell,-m,theta,phi,hp_mode_mm,hc_mode_mm)

        hp_mode, hc_mode = self.evaluate_on_sphere(ell,m,theta,phi,hp_mode,hc_mode)

        if mode_sum:
          hp_full = hp_full + hp_mode
          hc_full = hc_full + hc_mode
          #if fake_neg_modes and m != 0:
          #  hp_full = hp_full + hp_mode_mm
          #  hc_full = hc_full + hc_mode_mm
        else:
          if len(modes_to_evaluate)==1:
            hp_full[:] = hp_mode[:]
            hc_full[:] = hc_mode[:]
          else: 
            hp_full[:,ii] = hp_mode[:]
            hc_full[:,ii] = hc_mode[:]
      else:
        warning_str = "Your mode (ell,m) = ("+str(ell)+","+str(m)+") is not available!"
        raise Warning(warning_str)

      
      ii+=1

    if mode_sum:
      return t_mode, hp_full, hc_full #assumes all mode's have same temporal grid
    else: # helpful to have (l,m) list for understanding mode evaluations
      return modes_to_evaluate, t_mode, hp_full, hc_full


  #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  def evaluate_on_sphere(self,ell,m,theta,phi,hp_mode,hc_mode):
    """evaluate on the sphere"""

    if theta is not None: 
      #if phi is None: phi = 0.0
      if phi is None: raise ValueError('phi must have a value')
      sYlm_value =  _sYlm(-2,ll=ell,mm=m,theta=theta,phi=phi)
      h = sYlm_value*(hp_mode + 1.0j*hc_mode)
      hp_mode = h.real
      hc_mode = h.imag

    return hp_mode, hc_mode

  #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  def evaluate_single_mode(self,q, M, dist, f_low, samples, samples_units,ell,m):
    """ light wrapper around single mode evaluator"""

    t_mode, hp_mode, hc_mode = self.single_mode_dict[(ell,m)](q, M, dist, None, f_low, samples,
                                                              samples_units,singlemode_call=False)

    return t_mode, hp_mode, hc_mode


  #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  def evaluate_single_mode_by_symmetry(self,q, M, dist, f_low, samples,samples_units,ell,m):
    """ evaluate m<0 mode from m>0 mode and relationship between these"""

    if m<0:
      t_mode, hp_mode, hc_mode = self.evaluate_single_mode(q, M, dist, f_low, samples,samples_units,ell,-m)
      hp_mode, hc_mode         = self._generate_minus_m_mode(hp_mode,hc_mode,ell,-m)
    else:
      raise ValueError('m must be negative.')

    return t_mode, hp_mode, hc_mode


  #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  def generate_mode_eval_list(self,ell=None,m=None,minus_m=False):
    """generate list of (ell,m) modes to evaluate for.

      1) ell=m=None: use all available model modes
      2) ell=NUM, m=None: all modes up to ell_max = NUM. unmodelled modes set to zero
      3) list of [ell], [m] pairs: only use modes (ell,m). unmodelled modes set to zero 
         ex: ell=[3,2] and m=[2,2] generates a (3,2) and (2,2) mode.

      These three options produce a list of (ell,m) modes.

      Set minus_m=True to generate m<0 modes from m>0 modes."""

    ### generate list of nonnegative m modes to evaluate for ###
    if ell is None and m is None:
      modes_to_eval = self.all_model_modes()
    elif m is None:
      LMax = ell
      modes_to_eval = []
      for L in range(2,LMax+1):
        for emm in range(0,L+1):
          modes_to_eval.append((L,emm))
    else: # neither pythonic nor fast
      #modes_to_eval = [(x, y) for x in ell for y in m]
      modes_to_eval = []
      for ii in range(len(ell)):
        modes_to_eval.append((ell[ii],m[ii]))

    ### if m<0 requested, build these from m>=0 list ###
    if minus_m:
      modes_to_eval = self._extend_mode_list_minus_m(modes_to_eval)

    return modes_to_eval

  #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  def sort_mode_list(self,mode_list):
    """sort modes as (2,-2), (2,-1), ..., (2,2), (3,-3),(3,-2)..."""

    from operator import itemgetter

    mode_list = sorted(mode_list, key=itemgetter(0,1))
    return mode_list


  #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  def all_model_modes(self,minus_m=False):
    """ from single mode keys deduce all available model modes.
        If minus_m=True, include (ell,-m) whenever (ell,m) is available ."""

    model_modes = [(ell,m) for ell,m in self.single_mode_dict.keys()] # Py2 list, Py3 iterator

    if minus_m:
      model_modes = self._extend_mode_list_minus_m(model_modes)

    return model_modes


  #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  def single_mode(self,mode):
    """ Returns a single-mode object for mode=(ell,m).
        This object stores information for the (ell,m)-mode surrogate"""
    return self.single_mode_dict[mode] 


  #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  def match_surrogate(self, t_ref,h_ref,q, M=None, dist=None, theta=None,\
                     t_ref_units='dimensionless',ell=None,m=None,fake_neg_modes=True,\
                     t_low_adj=.0125,t_up_adj=.0125,speed='slow'):
    """ match discrete complex polarization (t_ref,h_ref) to surrogate waveform for 
        given input values. Inputs have same meaning as those passed to __call__

        Minimization (i.e. match) over time shifts and z-axis rotations"""

    # TODO: routine only works for hp,hc evaluated on the sphere. should extend to modes

    if (M is None and t_ref_units=='mks') or (M is not None and t_ref_units=='dimensionless'):
      raise ValueError('surrogate evaluations and reference temporal grid are inconsistent')


    if speed == 'slow': # repeated calls to surrogate evaluation routines

      ### setup minimization problem -- deduce common time grid and approximate minimization solution from discrete waveform ###
      time_sur,hp,hc = self.__call__(q=q,M=M,dist=dist,theta=theta,phi=0.0,\
                                  samples=self.time_all_modes(),samples_units='dimensionless',ell=ell,m=m,fake_neg_modes=fake_neg_modes)
      h_sur =  hp + 1.0j*hc

      # TODO: this deltaPhi is overall phase shift -- NOT a good guess for minimizations
      junk1, h2_eval, common_times, deltaT, deltaPhi = \
         _gwtools.setup_minimization_from_discrete_waveforms(time_sur,h_sur,t_ref,h_ref,t_low_adj,t_up_adj)

      ### (tc,phic)-parameterized waveform function to induce a parameterized norm ###
      def parameterized_waveform(x):
        tc   = x[0]
        phic = x[1]
        times = _gwtools.coordinate_time_shift(common_times,tc)        
        times,hp,hc = self.__call__(q=q,M=M,dist=dist,theta=theta,phi=phic,\
                                  samples=times,samples_units=t_ref_units,ell=ell,m=m,fake_neg_modes=fake_neg_modes)
        return hp + 1.0j*hc

    elif speed == 'fast': # build spline interpolant of modes, evaluate the interpolant

      modes_to_evaluate, t_mode, hp_full, hc_full = self.__call__(q=q,M=M,dist=dist,ell=ell,m=m,mode_sum=False,fake_neg_modes=fake_neg_modes)
      h_sphere = _gwtools.h_sphere_builder(modes_to_evaluate, hp_full, hc_full, t_mode)

      hp,hc=h_sphere(t_mode,theta=theta, phi=0.0, z_rot=None, psi_rot=None)
      h1=hp+1.0j*hc

      junk1, h2_eval, common_times, deltaT, deltaPhi = \
          _gwtools.setup_minimization_from_discrete_waveforms(t_mode,h1,t_ref,h_ref,t_low_adj,t_up_adj)
      parameterized_waveform = _gwtools.generate_parameterize_waveform(common_times,h_sphere,'h_sphere',theta)

    else:
      raise ValueError('not coded yet')

    ### solve minimization problem ###
    [guessed_norm,min_norm], opt_solution, hsur_align = \
      _gwtools.minimize_waveform_match(parameterized_waveform,\
                                      h2_eval,_gwtools.euclidean_norm_sqrd,\
                                      [deltaT,-deltaPhi/2.0],'nelder-mead')


    return min_norm, opt_solution, [common_times, hsur_align, h2_eval]


  #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  def h_sphere_builder(self, q, M=None,dist=None,ell=None,m=None):
    """Returns a function for evaluations of h(t,theta,phi;q,M,d) which include 
       all available modes. 

       This new function h(t,theta,phi;q,M,d)
       can be evaluated for rotations about z-axis and at any set of 
       points on the sphere. modes_to_evalute are also returned"""

    modes_to_evaluate, t_mode, hp_full, hc_full = \
      self(q=q, M=M, dist=dist, mode_sum=False,ell=ell,m=m)

    h_sphere = _gwtools.h_sphere_builder(modes_to_evaluate, hp_full, hc_full, t_mode)
    
    return h_sphere, modes_to_evaluate,


  #### below here are "private" member functions ###
  # These routine's carry out inner workings of multimode surrogate
  # class (such as memory allocation)

  #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  def _allocate_output_array(self,samples,num_modes,mode_sum):
    """ allocate memory for result of hp, hc.

    Input
    =====
    samples   --- array of time samples. None if using default
    num_modes --- number of harmonic modes (cols). set to 1 if summation over modes
    mode_sum  --- whether or not modes will be summed over (see code for why necessary)"""


    # Determine the number of time samples #
    if (samples is not None):
      sample_size = samples.shape[0]
    else:
      sample_size = self.time_all_modes().shape[0]

    # TODO: should the dtype be complex?
    if(num_modes==1): # return as vector instead of array
      hp_full = np.zeros(sample_size)
      hc_full = np.zeros(sample_size)
    else:
      hp_full = np.zeros((sample_size,num_modes))
      hc_full = np.zeros((sample_size,num_modes))

    return hp_full, hc_full


  #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  def _generate_minus_m_mode(self,hp_mode,hc_mode,ell,m):
    """ For m>0 positive modes hp_mode,hc_mode use h(l,-m) = (-1)^l h(l,m)^*
        to compute the m<0 mode. 

  See Eq. 78 of Kidder,Physical Review D 77, 044016 (2008), arXiv:0710.0614v1 [gr-qc]."""

    if (m<=0):
      raise ValueError('m must be nonnegative. m<0 will be generated for you from the m>0 mode.')
    else:
      hp_mode =   np.power(-1,ell) * hp_mode
      hc_mode = - np.power(-1,ell) * hc_mode

    return hp_mode, hc_mode


  #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  def _extend_mode_list_minus_m(self,mode_list):
    """ from list of [(ell,m)] pairs return a new list which includes m<0 too."""

    positive_modes = list(mode_list)
    for ell,m in positive_modes:
      if m>0:
        mode_list.append((ell,-m))
      if m<0:
        raise ValueError('your list already has negative modes!')

    return mode_list


####################################################
# TODO: loop over all data defined in IO class
def CompareSingleModeSurrogate(sur1,sur2):
  """ Compare data defining two surrogates"""

  agrees = []
  different = []
  no_check = []

  for key in sur1.__dict__.keys(): # Py2 list, Py3 iterator

    result = "Checking attribute %s... "%str(key)

    # surrogate data (ie numbers or array of numbers) 
    if key in ['B','V','R','fitparams_phase','fitparams_amp',\
               'fitparams_norm','greedy_points','eim_indices',\
               'time_info','fit_interval','tmin','tmax',\
               'modeled_data','fits_required','dt','times',\
               'fit_min','fit_max']:

      if np.max(np.abs(sur1.__dict__[key] - sur2.__dict__[key])) != 0:
        different.append(key)
      else:
        agrees.append(key)

    # surrogate tags (ie strings)
    elif key in ['fit_type_phase','fit_type_amp','fit_type_norm',\
                 'parameterization','affine_map','surrogate_mode_type',
                 't_units','surrogate_units','norms']:

      if sur1.__dict__[key] == sur2.__dict__[key]:
        agrees.append(key)
      else:
        different.append(key)

    else:
      no_check.append(key)

  print("Agrees:")
  print(agrees)
  print("Different:")
  print(different)
  print("Did not check:")
  print(no_check)






#!/usr/bin/env python

import numpy as np
import os
import unittest

if __package__ is "" or "None": # py2 and py3 compatible 
  print("setting __package__ to gwsurrogate.new so relative imports work")
  __package__="gwsurrogate.new"
from .nodeFunction import DummyNodeFunction, Polyfit1D
from gwsurrogate import parametric_funcs as pf

TEST_FILE = 'test.h5' # Gets created and deleted


def _tear_down():
    if os.path.isfile(TEST_FILE):
        os.remove(TEST_FILE)

def _set_up():
    # Don't overwrite this in case it's actually needed by something else
    if os.path.isfile(TEST_FILE):
        raise Exception("{} already exists! Please move or remove it.")


class BaseTest(unittest.TestCase):

    def setUp(self):
        _set_up()

    def tearDown(self):
        _tear_down()


class DummyNodeTester(BaseTest):

    def _verify_output(self, dnf, inputs, return_val):
        for x in inputs:
            y = dnf(x)
            if return_val is None:
                y2 = np.mean(x)
                self.assertEqual(y, y2)
            else:
                self.assertEqual(y, return_val)

    def _test(self, inputs, return_val=None):
        dnf = DummyNodeFunction(return_value=return_val)
        self._verify_output(dnf, inputs, return_val)
        dnf.save(TEST_FILE)

        dnf2 = DummyNodeFunction()
        dnf2.load(TEST_FILE)
        self._verify_output(dnf, inputs, return_val)
        _tear_down()

    def test_dummy_node_function(self):
        inputs = [[1.0], [2.0], np.array([3, 4.0]), np.ones((2, 3))]
        self._test(inputs, None)
        self._test(inputs, 1.2)


class Polyfit1DTester(BaseTest):

    def _verify_output(self, pf1d, inputs, outputs):
        for x, y in zip(inputs, outputs):
            yfit = pf1d([x])
            self.assertEqual(y, yfit)

    def _test(self, inputs, n_coefs, function_name):
        coefs = np.random.random(n_coefs)
        pf1d = Polyfit1D(function_name, coefs)
        func = pf.function_dict[function_name]
        outputs = [func(coefs, x) for x in inputs]
        self._verify_output(pf1d, inputs, outputs)

        pf1d.save(TEST_FILE)
        pf1d2 = Polyfit1D()
        pf1d2.load(TEST_FILE)
        self._verify_output(pf1d2, inputs, outputs)
        _tear_down()

    def test_all_functions(self):
        inputs = [0.1, 0.5]
        self._test(inputs, 1, 'polyval_1d')
        self._test(inputs, 5, 'polyval_1d')
        self._test(inputs, 3, 'ampfitfn1_1d')
        self._test(inputs, 3, 'ampfitfn2_1d')
        self._test(inputs, 4, 'phifitfn1_1d')
        #self._test(inputs, 5, 'ampfitfn3_1d')
        self._test(inputs, 4, 'ampfitfn4_1d')
        self._test(inputs, 5, 'nuSingularPlusPolynomial')
        self._test(inputs, 5, 'nuSingular2TermsPlusPolynomial')



""" catalog of parametric fitting functions """

from __future__ import division # for python 2

__copyright__    = "Copyright (C) 2014 Scott Field and Chad Galley"
__email__        = "sfield@astro.cornell.edu, crgalley@tapir.caltech.edu"
__status__       = "testing"
__author__       = "Jonathan Blackman, Scott Field, Chad Galley"
__contributors__ = []

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

import numpy as np
from .gwtools import gwtools as gwtools # from the package gwtools, import the module gwtools (gwtools.py)....



#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
def polyval_1d(coeffs,x):
  """ 1D polynomial defined by coeffs vector and evaluated 
  as numpy.polyval(coeffs,x)"""
  return np.polyval(coeffs, x)

#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
def ampfitfn1_1d(coeffs,x):
  """ PN inspired ampitude fit a0 + a1*nu**a2"""

  a0, a1, a2 = coeffs[:3]
  nu = gwtools.q_to_nu(x)
  return a0 + a1*nu**a2

#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
def ampfitfn2_1d(coeffs, x):
  """ PN inspired amplitude fit a0 + a1*np.abs(0.25-nu)**0.5 + a2*np.log(nu/0.25)"""

  a0, a1, a2 = coeffs[:3]
  nu = gwtools.q_to_nu(x)
  return a0 + a1*np.abs(0.25-nu)**0.5 + a2*np.log(nu/0.25)

#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
def phifitfn1_1d(coeffs,x):
  """ PN inspired phase fit a0 + a1*nu + a2*nu**2 + a3*np.log(nu)"""

  a0, a1, a2, a3 = coeffs[:4]
  nu = gwtools.q_to_nu(x)
  return a0 + a1*nu + a2*nu**2 + a3*np.log(nu)

#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
def ampfitfn3_1d(coeffs, x):
  """ PN inspired amplitude fit plus polynomial,
      a0*np.sqrt(1 - x) + a1*np.log(x) + a2(1 - x) + ... + aN*(1-x)**(N-1)"""

  a0 = coeffs[-1]
  a1 = coeffs[-2]

  polyCoefs = [c for c in coeffs[:-2]]
  polyCoefs.append(0.)

  return a0*np.sqrt(1. - x) + a1*np.log(x) + np.polyval(polyCoefs,1. - x)

#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
def ampfitfn4_1d(coeffs, x):
  """ a0*sqrt(1.0-x) + a1*(1.0-x) + a2*(1.0-x)^2 + a3*(1.0-x)^3"""

  a0, a1, a2, a3 = coeffs[:4]
  return a0*np.sqrt(1.0-x) + a1*(1.0-x) + a2*np.power(1.0-x,2) + a3*np.power(1.0-x,3)

#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
def ampfitfn5_1d(coeffs,x):
  """a0*np.sqrt(1 - x) + a1(1 - x) + ... + aN*(1-x)**N"""
  a0 = coeffs[-1]
  polyCoefs = [c for c in coeffs[:-1]]
  polyCoefs.append(0.)
  
  return a0*np.sqrt(1. - x) + np.polyval(polyCoefs,1. - x)

#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
def ampfitfn6_1d(coeffs,x):
  """a0*np.sqrt(1 - x) + a1*(1 - x)**1.5 + [ a2*(1 - x) + ... + aN*(1-x)**(N-1)"""
  a0 = coeffs[-1]
  a1 = coeffs[-2]
  polyCoefs = [c for c in coeffs[:-2]]
  polyCoefs.append(0.)

  return a0*np.sqrt(1. - x) + a1*(1. - x)**1.5 + np.polyval(polyCoefs,1. - x)

### these are for switching from (q,M) to surrogate's parameterization ###
def q_to_q(q):
  return q

def q_to_nu(q):
  return gwtools.q_to_nu(q)

### dictionary of fitting functions ###
function_dict = {
                 "polyval_1d": polyval_1d,
                 "ampfitfn1_1d": ampfitfn1_1d,
                 "ampfitfn2_1d": ampfitfn2_1d,
                 "ampfitfn4_1d": ampfitfn4_1d,
                 "phifitfn1_1d": phifitfn1_1d,
                 "nuSingularPlusPolynomial": ampfitfn5_1d,
                 "nuSingular2TermsPlusPolynomial": ampfitfn6_1d,
                 "q_to_q": q_to_q,
                 "q_to_nu":q_to_nu
                 }



"""GWSurrogate
   ===========

Provides 
  1. An to use easy interface to gravitational wave surrogate models
  2. Up-to-date database of surrogates and tools for downloading them
  3. Helper routines required for basic data analysis tasks


Example usage
-------------

To plot the EOBNRv2 surrogate included with this package (others available for download)

  ipmort gwsurrogate as gw
  EOB = gws.EvaluateSurrogate('gwsurrogate/tutorial/EOBNRv2_example/EOBNRv2_q1_2_NoSpin_SingleModes/l2_m2_len12239M_SurID19poly/')
  EOB.plot_sur(q_eval = 1.3)

Additional examples can be found in the accompanying ipython notebooks.


Surrogate data format
---------------------

Both HDF5 and text based surruogates assume 

1) Gravitational waves are written as: 

        h = h_+ + i h_x = Ae^{i phi} 

2) The following surrogate data is available:

        SurrogateID (string, repeat parent directory name)
	B (complex matrix, basis-by-times)
	eim_indices (array of integers, labeled from 0)
	greedy_points (ordered by their selection)
	tmin, tmax, dt. (Text surrogates store these in time_info.txt)
	qmin_fit, qmax_fit (range used for fitting).
	affine_map (Boolean; whether affine map to reference interval was 
                    used; use as flag for evaluating fits)
	fit_coeff_amp 
	fit_coeff_phase (coefficients of fitting functions; eim_indices-by-coefficients)
	V (not its inverse; to reconstruct the the orthonormal reduced basis E = B V)
	R (matrix of coefficients relating basis and waveforms, H = E R)

3) Each surrogate model is defined by a folder name (text) 

        MODELNAME_q[QMIN]_[QMAX]_NoSpin_[Multi(Single)]Mode/l[ELL]_m[M]_len[TIME]M_SurID[RBDIM]poly

or file name (hdf5)  

        ADD ME HERE!!!!

where, for example, MODELNAME = "EOBNRv2" and []-quantities are determined by the surrogate's setting

4) metadata.txt or info.data contains a detailed description of the surrogate model


surrogate.py
------------

Defines the classes HDF5Surrogate and TextSurrogate. These are low-level 
classes for loading, evaluating and plotting surrogate models (stored 
as text or hdf5 data files) as they are exported from the surrogate building 
code; these surrogates are limited to a fixed sampling rate and are dimensionless.

The EvaluateSurrogate class can be used to generate astrophysical surrogates
depending on the masses, distance to the source another other parameters 
of interest.

"""

__copyright__ = "Copyright (C) 2014 Scott Field, Chad Galley"
__email__ = "sfield@umd.edu"
__status__ = "testing"
__author__ = "Scott Field, Chad Galley"
__contributors__ = [
    # Alphabetical by first name.
    "Jonathan Blackman"]
__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

from .surrogate import *
from . import surrogate_repo



'''This file defines constants used by romgw, gwsurrogate and gwtools'''

'''import sys
if hasattr(sys, '_called_from_test'):
  print "Called from within a test run"
  # older constants (outdated 7/9/2015) -- keep for regression testing
  c         = 2.99792458e8     # Speed of light (MKS)
  G         = 6.67428e-11      # Gravitation constant (MKS)
  Msun      = 1.98892e30       # Solar mass (kg)
  Msuninsec = Msun * G / c**3  # Solar mass (secs)
  Mpcinm    = 3.08568025e22    # Megaparsecs (m)
  Mpcinkm   = Mpcinm/1000.0    # Megaparsecs (km)

else:
  print "Using LAL values for constants"
  # These contants are found in LAL code
  c         = 299792458.0
  G         = 6.67384e-11
  Msun      = 1.9885469549614615e+30
  Msuninsec = Msun * G / c**3 # TODO: check the product obtains ~11 digits of accuracy
  Mpcinm    = 3.085677581491367e+22
  Mpcinkm   = Mpcinm/1000.0    # Megaparsecs (km)
  Msuninkg = Msun'''

# romgw constants, and older gwsurroate constants # 
# TODO: consider keeping these in favor of lal constants
# commented out above. The codes will reproduce previous results.
c         = 2.99792458e8     # Speed of light (MKS)
G         = 6.67428e-11      # Gravitation constant (MKS)
Msun      = 1.98892e30       # Solar mass (kg)

# TODO: Msuninsec should be hardcoded -- its known to more digits than G or Msum
Msuninsec = Msun * G / c**3  # Solar mass (secs)

Mpcinm    = 3.08568025e22    # Megaparsecs (m)
Mpcinkm   = Mpcinm/1000.0    # Megaparsecs (km)
Msuninkg  = Msun             # for romgw 

"""Euler-Mascheroni constant"""
gamma_E = 0.577215664901532


# Jonathan's constants
#Msun = 1.9891e30
#Mpcinm = 3.08567758e22
#SPEED_OF_LIGHT_C = 299792458
#GRAVITATIONAL_CONSTANT_G = 6.67384e-11
#G = 6.67384e-11
#Msuninsec = Msun * G / c**3
#Mpcinkm   = Mpcinm/1000.0    # Megaparsecs (km)




""" Surrogate download tool """

from __future__ import division # for python 2


__copyright__ = "Copyright (C) 2014 Scott Field and Chad Galley"
__email__     = "sfield@astro.cornell.edu, crgalley@tapir.caltech.edu"
__status__    = "testing"
__author__    = "Scott Field, Chad Galley"

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

import os
from collections import namedtuple

### Naming convention: dictionary KEY should match file name KEY.tar.gz ###
surrogate_info = namedtuple('surrogate_info', ['url', 'desc'])

### dictionary of all known surrogates ###
_surrogate_world = {}

_surrogate_world['EOBNRv2'] = \
  surrogate_info('https://www.dropbox.com/s/uyliuy37uczu3ug/EOBNRv2.tar.gz',\
                 '''(WARNING: older surrogate may not evaluate with 
most recent version of gwsurrogate) Collection of single mode surrogates 
from mass ratios 1 to 10, as long as 190000M and modes (2,1), (2,2), (3,3), 
(4,4), (5,5). This is not a true multi-mode surrogate, and relative time/phase
information between the modes have not been preseved. For more information see 
http://journals.aps.org/prx/abstract/10.1103/PhysRevX.4.031006''')

_surrogate_world['SpEC_q1_10_NoSpin'] = \
  surrogate_info('https://www.black-holes.org/surrogates/data/SpEC_q1_10_NoSpin_nu5thDegPoly_exclude_2_0.h5',\
                 '''A multimode surrogate model built from numerical relativity
simulations performed with SpEC. The surrogate covers mass ratios from 1 to 10,
durations corresponding to about 15 orbits before merger, and many harmonic
modes. For more information see https://www.black-holes.org/surrogates/ or
http://arxiv.org/abs/1502.07758''')

def download_path():
  '''return the default path for downloaded surrogates'''

  import gwsurrogate
  import os
  gws_path = os.path.dirname(gwsurrogate.__file__)
  return gws_path+'/../surrogate_downloads/'

def list():
  '''show all known surrogates available for download'''

  for surr_key in _surrogate_world.keys():
    print(surr_key+'...')
    print('  url: '+_surrogate_world[surr_key].url)
    print("  Description: "+_surrogate_world[surr_key].desc+'\n\n')

def _unzip(surr_name,sdir=download_path()):
  '''unzip a tar.gz surrogate and remove the tar.gz file'''

  os.chdir(sdir)
  os.system('tar -xvzf '+surr_name)
  os.remove(surr_name)

  return sdir+surr_name.split('.')[0]

def get(surr_name,sdir=download_path()):
  '''pass a valid surr_name from the repo list and download location sdir. 
     The default path is used if no location supplied. tar.gz surrogates
     are automatically unziped. The new surrogate path is returned.'''

  if surr_name in _surrogate_world:
    surr_url = _surrogate_world[surr_name].url
    os.system('wget --directory-prefix='+sdir+' '+surr_url)
  else:
    raise ValueError("No surrogate package exits")

  # deduce the surrogte file name and extension type
  # one can directly load a surrogate frmo surr_path
  file_name = surr_url.split('/')[-1]
  if file_name.split('.')[1] == 'tar': # assumed to be *.h5 or *.tar.gz
    surr_path = _unzip(file_name,sdir)
  else:
    surr_path = sdir+file_name

  return surr_path




import sys

try:
  from setuptools import setup
  setup
except ImportError: # currently not supported
  raise ImportError("GWSurrogate requires setuptools to correctly install the gwools submodule")
  #from distutils.core import setup # currently not supported
  #setup

# To render markdown. See https://github.com/pypa/pypi-legacy/issues/148
try:
    import pypandoc
    long_description = pypandoc.convert('README.md', 'rst')
except ImportError:
    long_description = open('README.md').read()


setup(name='gwsurrogate',
      version='0.6.3',
      author='Jonathan Blackman, Scott Field, Chad Galley',
      author_email='sfield@astro.cornell.edu',
      packages=['gwsurrogate'],
      license='MIT',
      include_package_data=True,
      contributors=[
      # Alphabetical by last name.
      ""],
      description='An easy to use interface to gravitational wave surrogate models',
      long_description=long_description,
      # will start new downloads if these are installed in a non-standard location
      # install_requires=["numpy","matplotlib","scipy"],
      classifiers=[
                'Intended Audience :: Other Audience',
                'Intended Audience :: Science/Research',
                'Natural Language :: English',
                'License :: OSI Approved :: MIT License',
                'Programming Language :: Python',
                'Topic :: Scientific/Engineering',
                'Topic :: Scientific/Engineering :: Mathematics',
                'Topic :: Scientific/Engineering :: Physics',
      ],
)
